from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Union

from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource, SQLSourceKind, TableIngestMixIn
from chalk.sql.protocols import SQLSourceWithTableIngestProtocol
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL


_MYSQL_HOST_NAME = "MYSQL_HOST"
_MYSQL_TCP_PORT_NAME = "MYSQL_TCP_PORT"
_MYSQL_DATABASE_NAME = "MYSQL_DATABASE"
_MYSQL_USER_NAME = "MYSQL_USER"
_MYSQL_PWD_NAME = "MYSQL_PWD"


class MySQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIngestProtocol):
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
        engine_args: Optional[Dict[str, Any]] = None,
        async_engine_args: Optional[Dict[str, Any]] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            import pymysql
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[mysql]")
        del pymysql
        self.name = name
        self.host = host or load_integration_variable(
            name=_MYSQL_HOST_NAME, integration_name=name, override=integration_variable_override
        )
        self.port = (
            int(port)
            if port is not None
            else load_integration_variable(
                name=_MYSQL_TCP_PORT_NAME, integration_name=name, parser=int, override=integration_variable_override
            )
        )
        self.db = db or load_integration_variable(
            name=_MYSQL_DATABASE_NAME, integration_name=name, override=integration_variable_override
        )
        self.user = user or load_integration_variable(
            name=_MYSQL_USER_NAME, integration_name=name, override=integration_variable_override
        )
        self.password = password or load_integration_variable(
            name=_MYSQL_PWD_NAME, integration_name=name, override=integration_variable_override
        )
        self.ingested_tables: Dict[str, Any] = {}
        if engine_args is None:
            engine_args = {}
        if async_engine_args is None:
            async_engine_args = {}
        engine_args.setdefault("pool_size", 20)
        engine_args.setdefault("max_overflow", 60)
        async_engine_args.setdefault("pool_size", 20)
        async_engine_args.setdefault("max_overflow", 60)
        # We set the default isolation level to autocommit since the SQL sources are read-only, and thus
        # transactions are not needed
        # Setting the isolation level on the engine, instead of the connection, avoids
        # a DBAPI statement to reset the transactional level back to the default before returning the
        # connection to the pool
        engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
        async_engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
        BaseSQLSource.__init__(self, name=name, engine_args=engine_args, async_engine_args=async_engine_args)

    kind = SQLSourceKind.mysql

    def get_sqlglot_dialect(self) -> str | None:
        return "mysql"

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="mysql+pymysql",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.db,
        )

    def async_local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="mysql+aiomysql",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.db,
        )

    def _recreate_integration_variables(self) -> Dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_MYSQL_HOST_NAME, self.name, self.host),
                create_integration_variable(_MYSQL_TCP_PORT_NAME, self.name, self.port),
                create_integration_variable(_MYSQL_DATABASE_NAME, self.name, self.db),
                create_integration_variable(_MYSQL_USER_NAME, self.name, self.user),
                create_integration_variable(_MYSQL_PWD_NAME, self.name, self.password),
            ]
            if v is not None
        }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/authtesting.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n!chalk/server/v1/authtesting.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto" \n\x1eGetUnauthedTestEndpointRequest"\x1e\n\x1cGetAuthedTestEndpointRequest"\x1e\n\x1cGetViewerTestEndpointRequest"%\n#GetDataScientistTestEndpointRequest"!\n\x1fGetDeveloperTestEndpointRequest"\x1d\n\x1bGetAdminTestEndpointRequest"\x1d\n\x1bGetOwnerTestEndpointRequest"!\n\x1fGetUnauthedTestEndpointResponse"\x1f\n\x1dGetAuthedTestEndpointResponse"\x1f\n\x1dGetViewerTestEndpointResponse"&\n$GetDataScientistTestEndpointResponse""\n GetDeveloperTestEndpointResponse"\x1e\n\x1cGetAdminTestEndpointResponse"\x1e\n\x1cGetOwnerTestEndpointResponse2\xb5\x07\n\x12\x41uthTestingService\x12\x84\x01\n\x17GetUnauthedTestEndpoint\x12/.chalk.server.v1.GetUnauthedTestEndpointRequest\x1a\x30.chalk.server.v1.GetUnauthedTestEndpointResponse"\x06\x90\x02\x01\x80}\x01\x12~\n\x15GetAuthedTestEndpoint\x12-.chalk.server.v1.GetAuthedTestEndpointRequest\x1a..chalk.server.v1.GetAuthedTestEndpointResponse"\x06\x90\x02\x01\x80}\x02\x12~\n\x15GetViewerTestEndpoint\x12-.chalk.server.v1.GetViewerTestEndpointRequest\x1a..chalk.server.v1.GetViewerTestEndpointResponse"\x06\x90\x02\x01\x80}\x0b\x12\x93\x01\n\x1cGetDataScientistTestEndpoint\x12\x34.chalk.server.v1.GetDataScientistTestEndpointRequest\x1a\x35.chalk.server.v1.GetDataScientistTestEndpointResponse"\x06\x90\x02\x01\x80}\x03\x12\x87\x01\n\x18GetDeveloperTestEndpoint\x12\x30.chalk.server.v1.GetDeveloperTestEndpointRequest\x1a\x31.chalk.server.v1.GetDeveloperTestEndpointResponse"\x06\x90\x02\x01\x80}\x11\x12{\n\x14GetAdminTestEndpoint\x12,.chalk.server.v1.GetAdminTestEndpointRequest\x1a-.chalk.server.v1.GetAdminTestEndpointResponse"\x06\x90\x02\x01\x80}\x0c\x12{\n\x14GetOwnerTestEndpoint\x12,.chalk.server.v1.GetOwnerTestEndpointRequest\x1a-.chalk.server.v1.GetOwnerTestEndpointResponse"\x06\x90\x02\x01\x80}\nB\x99\x01\n\x13\x63om.chalk.server.v1B\x10\x41uthtestingProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.authtesting_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\020AuthtestingProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_AUTHTESTINGSERVICE"].methods_by_name["GetUnauthedTestEndpoint"]._options = None
    _globals["_AUTHTESTINGSERVICE"].methods_by_name[
        "GetUnauthedTestEndpoint"
    ]._serialized_options = b"\220\002\001\200}\001"
    _globals["_AUTHTESTINGSERVICE"].methods_by_name["GetAuthedTestEndpoint"]._options = None
    _globals["_AUTHTESTINGSERVICE"].methods_by_name[
        "GetAuthedTestEndpoint"
    ]._serialized_options = b"\220\002\001\200}\002"
    _globals["_AUTHTESTINGSERVICE"].methods_by_name["GetViewerTestEndpoint"]._options = None
    _globals["_AUTHTESTINGSERVICE"].methods_by_name[
        "GetViewerTestEndpoint"
    ]._serialized_options = b"\220\002\001\200}\013"
    _globals["_AUTHTESTINGSERVICE"].methods_by_name["GetDataScientistTestEndpoint"]._options = None
    _globals["_AUTHTESTINGSERVICE"].methods_by_name[
        "GetDataScientistTestEndpoint"
    ]._serialized_options = b"\220\002\001\200}\003"
    _globals["_AUTHTESTINGSERVICE"].methods_by_name["GetDeveloperTestEndpoint"]._options = None
    _globals["_AUTHTESTINGSERVICE"].methods_by_name[
        "GetDeveloperTestEndpoint"
    ]._serialized_options = b"\220\002\001\200}\021"
    _globals["_AUTHTESTINGSERVICE"].methods_by_name["GetAdminTestEndpoint"]._options = None
    _globals["_AUTHTESTINGSERVICE"].methods_by_name[
        "GetAdminTestEndpoint"
    ]._serialized_options = b"\220\002\001\200}\014"
    _globals["_AUTHTESTINGSERVICE"].methods_by_name["GetOwnerTestEndpoint"]._options = None
    _globals["_AUTHTESTINGSERVICE"].methods_by_name["GetOwnerTestEndpoint"]._serialized_options = b"\220\002\001\200}\n"
    _globals["_GETUNAUTHEDTESTENDPOINTREQUEST"]._serialized_start = 87
    _globals["_GETUNAUTHEDTESTENDPOINTREQUEST"]._serialized_end = 119
    _globals["_GETAUTHEDTESTENDPOINTREQUEST"]._serialized_start = 121
    _globals["_GETAUTHEDTESTENDPOINTREQUEST"]._serialized_end = 151
    _globals["_GETVIEWERTESTENDPOINTREQUEST"]._serialized_start = 153
    _globals["_GETVIEWERTESTENDPOINTREQUEST"]._serialized_end = 183
    _globals["_GETDATASCIENTISTTESTENDPOINTREQUEST"]._serialized_start = 185
    _globals["_GETDATASCIENTISTTESTENDPOINTREQUEST"]._serialized_end = 222
    _globals["_GETDEVELOPERTESTENDPOINTREQUEST"]._serialized_start = 224
    _globals["_GETDEVELOPERTESTENDPOINTREQUEST"]._serialized_end = 257
    _globals["_GETADMINTESTENDPOINTREQUEST"]._serialized_start = 259
    _globals["_GETADMINTESTENDPOINTREQUEST"]._serialized_end = 288
    _globals["_GETOWNERTESTENDPOINTREQUEST"]._serialized_start = 290
    _globals["_GETOWNERTESTENDPOINTREQUEST"]._serialized_end = 319
    _globals["_GETUNAUTHEDTESTENDPOINTRESPONSE"]._serialized_start = 321
    _globals["_GETUNAUTHEDTESTENDPOINTRESPONSE"]._serialized_end = 354
    _globals["_GETAUTHEDTESTENDPOINTRESPONSE"]._serialized_start = 356
    _globals["_GETAUTHEDTESTENDPOINTRESPONSE"]._serialized_end = 387
    _globals["_GETVIEWERTESTENDPOINTRESPONSE"]._serialized_start = 389
    _globals["_GETVIEWERTESTENDPOINTRESPONSE"]._serialized_end = 420
    _globals["_GETDATASCIENTISTTESTENDPOINTRESPONSE"]._serialized_start = 422
    _globals["_GETDATASCIENTISTTESTENDPOINTRESPONSE"]._serialized_end = 460
    _globals["_GETDEVELOPERTESTENDPOINTRESPONSE"]._serialized_start = 462
    _globals["_GETDEVELOPERTESTENDPOINTRESPONSE"]._serialized_end = 496
    _globals["_GETADMINTESTENDPOINTRESPONSE"]._serialized_start = 498
    _globals["_GETADMINTESTENDPOINTRESPONSE"]._serialized_end = 528
    _globals["_GETOWNERTESTENDPOINTRESPONSE"]._serialized_start = 530
    _globals["_GETOWNERTESTENDPOINTRESPONSE"]._serialized_end = 560
    _globals["_AUTHTESTINGSERVICE"]._serialized_start = 563
    _globals["_AUTHTESTINGSERVICE"]._serialized_end = 1512
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/topic.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1b\x63halk/server/v1/topic.proto\x12\x0f\x63halk.server.v1"\'\n\x08SQSTopic\x12\x1b\n\tqueue_url\x18\x01 \x01(\tR\x08queueUrl"G\n\x0bPubSubTopic\x12\x1d\n\nproject_id\x18\x01 \x01(\tR\tprojectId\x12\x19\n\x08topic_id\x18\x02 \x01(\tR\x07topicId"\x8d\x01\n\x05Topic\x12\x38\n\tsqs_topic\x18\x01 \x01(\x0b\x32\x19.chalk.server.v1.SQSTopicH\x00R\x08sqsTopic\x12\x41\n\x0cpubsub_topic\x18\x02 \x01(\x0b\x32\x1c.chalk.server.v1.PubSubTopicH\x00R\x0bpubsubTopicB\x07\n\x05topicB\x93\x01\n\x13\x63om.chalk.server.v1B\nTopicProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.topic_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\nTopicProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_SQSTOPIC"]._serialized_start = 48
    _globals["_SQSTOPIC"]._serialized_end = 87
    _globals["_PUBSUBTOPIC"]._serialized_start = 89
    _globals["_PUBSUBTOPIC"]._serialized_end = 160
    _globals["_TOPIC"]._serialized_start = 163
    _globals["_TOPIC"]._serialized_end = 304
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.chart_pb2 import (
    GetChartSnapshotRequest,
    GetChartSnapshotResponse,
    ListChartsRequest,
    ListChartsResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class ChartsServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    ListCharts: UnaryUnaryMultiCallable[
        ListChartsRequest,
        ListChartsResponse,
    ]
    GetChartSnapshot: UnaryUnaryMultiCallable[
        GetChartSnapshotRequest,
        GetChartSnapshotResponse,
    ]

class ChartsServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def ListCharts(
        self,
        request: ListChartsRequest,
        context: ServicerContext,
    ) -> ListChartsResponse: ...
    @abstractmethod
    def GetChartSnapshot(
        self,
        request: GetChartSnapshotRequest,
        context: ServicerContext,
    ) -> GetChartSnapshotResponse: ...

def add_ChartsServiceServicer_to_server(servicer: ChartsServiceServicer, server: Server) -> None: ...



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class Series(_message.Message):
    __slots__ = ("points", "label", "units")
    POINTS_FIELD_NUMBER: _ClassVar[int]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    UNITS_FIELD_NUMBER: _ClassVar[int]
    points: _containers.RepeatedScalarFieldContainer[float]
    label: str
    units: str
    def __init__(
        self, points: _Optional[_Iterable[float]] = ..., label: _Optional[str] = ..., units: _Optional[str] = ...
    ) -> None: ...

class Chart(_message.Message):
    __slots__ = ("title", "series", "x_timestamp_ms")
    TITLE_FIELD_NUMBER: _ClassVar[int]
    SERIES_FIELD_NUMBER: _ClassVar[int]
    X_TIMESTAMP_MS_FIELD_NUMBER: _ClassVar[int]
    title: str
    series: _containers.RepeatedCompositeFieldContainer[Series]
    x_timestamp_ms: _containers.RepeatedScalarFieldContainer[int]
    def __init__(
        self,
        title: _Optional[str] = ...,
        series: _Optional[_Iterable[_Union[Series, _Mapping]]] = ...,
        x_timestamp_ms: _Optional[_Iterable[int]] = ...,
    ) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.server.v1 import webhook_pb2 as chalk_dot_server_dot_v1_dot_webhook__pb2


class WebhookServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateWebhook = channel.unary_unary(
            "/chalk.server.v1.WebhookService/CreateWebhook",
            request_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.CreateWebhookRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.CreateWebhookResponse.FromString,
        )
        self.UpdateWebhook = channel.unary_unary(
            "/chalk.server.v1.WebhookService/UpdateWebhook",
            request_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.UpdateWebhookRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.UpdateWebhookResponse.FromString,
        )
        self.DeleteWebhook = channel.unary_unary(
            "/chalk.server.v1.WebhookService/DeleteWebhook",
            request_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.DeleteWebhookRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.DeleteWebhookResponse.FromString,
        )
        self.GetWebhook = channel.unary_unary(
            "/chalk.server.v1.WebhookService/GetWebhook",
            request_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.GetWebhookRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.GetWebhookResponse.FromString,
        )
        self.ListWebhooks = channel.unary_unary(
            "/chalk.server.v1.WebhookService/ListWebhooks",
            request_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.ListWebhooksRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.ListWebhooksResponse.FromString,
        )


class WebhookServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def CreateWebhook(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdateWebhook(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def DeleteWebhook(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetWebhook(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListWebhooks(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_WebhookServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "CreateWebhook": grpc.unary_unary_rpc_method_handler(
            servicer.CreateWebhook,
            request_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.CreateWebhookRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.CreateWebhookResponse.SerializeToString,
        ),
        "UpdateWebhook": grpc.unary_unary_rpc_method_handler(
            servicer.UpdateWebhook,
            request_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.UpdateWebhookRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.UpdateWebhookResponse.SerializeToString,
        ),
        "DeleteWebhook": grpc.unary_unary_rpc_method_handler(
            servicer.DeleteWebhook,
            request_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.DeleteWebhookRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.DeleteWebhookResponse.SerializeToString,
        ),
        "GetWebhook": grpc.unary_unary_rpc_method_handler(
            servicer.GetWebhook,
            request_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.GetWebhookRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.GetWebhookResponse.SerializeToString,
        ),
        "ListWebhooks": grpc.unary_unary_rpc_method_handler(
            servicer.ListWebhooks,
            request_deserializer=chalk_dot_server_dot_v1_dot_webhook__pb2.ListWebhooksRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_webhook__pb2.ListWebhooksResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.server.v1.WebhookService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class WebhookService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def CreateWebhook(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.WebhookService/CreateWebhook",
            chalk_dot_server_dot_v1_dot_webhook__pb2.CreateWebhookRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_webhook__pb2.CreateWebhookResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdateWebhook(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.WebhookService/UpdateWebhook",
            chalk_dot_server_dot_v1_dot_webhook__pb2.UpdateWebhookRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_webhook__pb2.UpdateWebhookResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def DeleteWebhook(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.WebhookService/DeleteWebhook",
            chalk_dot_server_dot_v1_dot_webhook__pb2.DeleteWebhookRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_webhook__pb2.DeleteWebhookResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetWebhook(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.WebhookService/GetWebhook",
            chalk_dot_server_dot_v1_dot_webhook__pb2.GetWebhookRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_webhook__pb2.GetWebhookResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListWebhooks(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.WebhookService/ListWebhooks",
            chalk_dot_server_dot_v1_dot_webhook__pb2.ListWebhooksRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_webhook__pb2.ListWebhooksResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from __future__ import annotations

import inspect
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Collection

from chalk.utils.duration import CronTab, Duration

if TYPE_CHECKING:
    from chalk.client.models import FeatureReference


class ScheduledQuery:
    def __init__(
        self,
        name: str,
        schedule: CronTab | Duration,
        output: Collection[FeatureReference],
        recompute_features: bool | Collection[FeatureReference] = True,
        max_samples: int | None = None,
        lower_bound: datetime | None = None,
        upper_bound: datetime | None = None,
        tags: Collection[str] | None = None,
        required_resolver_tags: Collection[str] | None = None,
        store_online: bool = True,
        store_offline: bool = True,
        incremental_resolvers: Collection[str] | None = None,
        planner_options: dict[str, str] | None = None,
        resource_group: str | None = None,
    ):
        """Create an offline query which runs on a schedule.

        Scheduled queries do not produce datasets, but persist their results in the
        online and/or offline feature stores.

        By default, scheduled queries use incrementalization to only ingest data that
        has been updated since the last run.

        Parameters
        ----------
        name
            A unique name for the scheduled query. The name of the scheduled query
            will show up in the dashboard and will be uset to set the incremetalization
            metadata.
        schedule
            A cron schedule or a `Duration` object representing the interval at which
            the query should run.
        output
            The features that this query will compute. Namespaces are exploded into all
            features in the namespace.
        recompute_features
            Whether to recompute all features or load from the feature store.
            If `True`, all features will be recomputed.
            If `False`, all features will be loaded from the feature store.
            If a list of features, only those features will be recomputed, and the rest
            will be loaded from the feature store.
        max_samples
            The maximum number of samples to compute.
        lower_bound
            A hard-coded lower bound for the query. If set, the query will not use
            incrementalization.
        upper_bound
            A hard-coded upper bound for the query. If set, the query will not use
            incrementalization.
        tags
            Allows selecting resolvers with these tags.
        required_resolver_tags
            Requires that resolvers have these tags.
        store_online
            Whether to store the results of this query in the online store.
        store_offline
            Whether to store the results of this query in the offline store.
        incremental_resolvers
            If set to None, Chalk will incrementalize resolvers in the query's root namespaces.
            If set to a list of resolvers, this set will be used for incrementalization.
            Incremental resolvers must return a feature time in its output, and must return a `DataFrame`.
            Most commonly, this will be the name of a SQL file resolver. Chalk will ingest all new data
            from these resolvers and propagate changes to values in the root namespace.
        planner_options
            A dictionary of options to pass to the planner. These are typically provided by Chalk Support
            for specific use cases.
        resource_group
            The resource group to use for the query. If not set, the default resource group will be used.


        Returns
        -------
        ScheduledQuery
            A scheduled query object.

        Examples
        --------
        >>> from chalk.queries import ScheduledQuery
        >>> # this scheduled query will automatically run every 5 minutes after `chalk apply`
        >>> ScheduledQuery(
        ...     name="ingest_users",
        ...     schedule="*/5 * * * *",
        ...     output=[User],
        ...     store_online=True,
        ...     store_offline=True,
        ... )
        """
        super().__init__()
        self.errors = []

        if name in CRON_QUERY_REGISTRY:
            self.errors.append(
                f"A scheduled query with name '{name}' already exists. Scheduled query names must be unique."
            )

        if len(output) == 0:
            self.errors.append(
                f"Scheduled query '{name}' was instantiated with an empty set of outputs. At least one output is required."
            )

        if lower_bound is not None:
            lower_bound = lower_bound.astimezone(tz=timezone.utc)
        if upper_bound is not None:
            upper_bound = upper_bound.astimezone(tz=timezone.utc)

        caller_filename = None
        frame = inspect.currentframe()
        assert frame is not None, "Failed to get current frame"
        caller_frame = frame.f_back
        assert caller_frame is not None, "Failed to get caller frame"
        caller_filename = caller_frame.f_code.co_filename
        del frame

        if not store_offline and not store_online:
            self.errors.append(
                f"Scheduled query '{name}' was instantiated with `store_offline=False` and `store_online=False`. Running it will have no effect, as it does not store any data."
            )

        self.name = name
        self.cron = schedule
        self.output = [str(f) for f in output]
        self.max_samples = max_samples
        self.recompute_features = (
            recompute_features
            if recompute_features is True or recompute_features is False
            else [str(f) for f in recompute_features]
        )
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.tags = tags
        self.required_resolver_tags = required_resolver_tags
        self.filename = caller_filename
        self.store_online = store_online
        self.store_offline = store_offline
        if incremental_resolvers is not None and isinstance(incremental_resolvers, str):
            self.errors.append(
                f"Scheduled query '{name}' was instantiated with `incremental_resolvers={incremental_resolvers}`, but `{incremental_resolvers}` must be a list of resolver names."
            )
        self.incremental_resolvers = incremental_resolvers
        self.planner_options = {k: str(v) for k, v in planner_options.items()} if planner_options else None
        self.resource_group = resource_group

        CRON_QUERY_REGISTRY[name] = self


CRON_QUERY_REGISTRY: dict[str, ScheduledQuery] = {}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/artifacts/v1/chart.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1e\x63halk/artifacts/v1/chart.proto\x12\x12\x63halk.artifacts.v1"\x86\x03\n\x0c\x41lertTrigger\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x41\n\x08severity\x18\x02 \x01(\x0e\x32%.chalk.artifacts.v1.AlertSeverityKindR\x08severity\x12P\n\x12threshold_position\x18\x03 \x01(\x0e\x32!.chalk.artifacts.v1.ThresholdKindR\x11thresholdPosition\x12\'\n\x0fthreshold_value\x18\x04 \x01(\x02R\x0ethresholdValue\x12$\n\x0bseries_name\x18\x05 \x01(\tH\x00R\nseriesName\x88\x01\x01\x12&\n\x0c\x63hannel_name\x18\x06 \x01(\tH\x01R\x0b\x63hannelName\x88\x01\x01\x12%\n\x0b\x64\x65scription\x18\x07 \x01(\tH\x02R\x0b\x64\x65scription\x88\x01\x01\x42\x0e\n\x0c_series_nameB\x0f\n\r_channel_nameB\x0e\n\x0c_description"K\n\x15\x44\x61tasetFeatureOperand\x12\x18\n\x07\x64\x61taset\x18\x01 \x01(\tR\x07\x64\x61taset\x12\x18\n\x07\x66\x65\x61ture\x18\x02 \x01(\tR\x07\x66\x65\x61ture"\xdb\x02\n\rMetricFormula\x12\x39\n\x04kind\x18\x01 \x01(\x0e\x32%.chalk.artifacts.v1.MetricFormulaKindR\x04kind\x12\x39\n\x16single_series_operands\x18\x02 \x01(\x03H\x00R\x14singleSeriesOperands\x88\x01\x01\x12\x32\n\x15multi_series_operands\x18\x03 \x03(\x03R\x13multiSeriesOperands\x12\x63\n\x18\x64\x61taset_feature_operands\x18\x04 \x01(\x0b\x32).chalk.artifacts.v1.DatasetFeatureOperandR\x16\x64\x61tasetFeatureOperands\x12\x17\n\x04name\x18\x05 \x01(\tH\x01R\x04name\x88\x01\x01\x42\x19\n\x17_single_series_operandsB\x07\n\x05_name"\x9c\x01\n\x0cMetricFilter\x12\x32\n\x04kind\x18\x01 \x01(\x0e\x32\x1e.chalk.artifacts.v1.FilterKindR\x04kind\x12\x42\n\ncomparator\x18\x02 \x01(\x0e\x32".chalk.artifacts.v1.ComparatorKindR\ncomparator\x12\x14\n\x05value\x18\x03 \x03(\tR\x05value"\xb7\x02\n\x12MetricConfigSeries\x12\x36\n\x06metric\x18\x01 \x01(\x0e\x32\x1e.chalk.artifacts.v1.MetricKindR\x06metric\x12:\n\x07\x66ilters\x18\x02 \x03(\x0b\x32 .chalk.artifacts.v1.MetricFilterR\x07\x66ilters\x12\x17\n\x04name\x18\x03 \x01(\tH\x00R\x04name\x88\x01\x01\x12O\n\x0fwindow_function\x18\x04 \x01(\x0e\x32&.chalk.artifacts.v1.WindowFunctionKindR\x0ewindowFunction\x12:\n\x08group_by\x18\x05 \x03(\x0e\x32\x1f.chalk.artifacts.v1.GroupByKindR\x07groupByB\x07\n\x05_name"\xab\x02\n\x0cMetricConfig\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12#\n\rwindow_period\x18\x02 \x01(\tR\x0cwindowPeriod\x12>\n\x06series\x18\x03 \x03(\x0b\x32&.chalk.artifacts.v1.MetricConfigSeriesR\x06series\x12=\n\x08\x66ormulas\x18\x04 \x03(\x0b\x32!.chalk.artifacts.v1.MetricFormulaR\x08\x66ormulas\x12:\n\x07trigger\x18\x05 \x01(\x0b\x32 .chalk.artifacts.v1.AlertTriggerR\x07trigger\x12\'\n\x0fgraph_generated\x18\x06 \x01(\x08R\x0egraphGenerated"\xee\x01\n\x05\x43hart\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x38\n\x06\x63onfig\x18\x02 \x01(\x0b\x32 .chalk.artifacts.v1.MetricConfigR\x06\x63onfig\x12\x42\n\x0b\x65ntity_kind\x18\x03 \x01(\x0e\x32!.chalk.artifacts.v1.ChartLinkKindR\nentityKind\x12 \n\tentity_id\x18\x04 \x01(\tH\x00R\x08\x65ntityId\x88\x01\x01\x12\'\n\x0fgraph_generated\x18\x05 \x01(\x08R\x0egraphGeneratedB\x0c\n\n_entity_id*\xc3\n\n\nMetricKind\x12\x1b\n\x17METRIC_KIND_UNSPECIFIED\x10\x00\x12%\n!METRIC_KIND_FEATURE_REQUEST_COUNT\x10\x01\x12\x1f\n\x1bMETRIC_KIND_FEATURE_LATENCY\x10\x02\x12!\n\x1dMETRIC_KIND_FEATURE_STALENESS\x10\x03\x12\x1d\n\x19METRIC_KIND_FEATURE_VALUE\x10\x04\x12\x1d\n\x19METRIC_KIND_FEATURE_WRITE\x10\x05\x12"\n\x1eMETRIC_KIND_FEATURE_NULL_RATIO\x10\x06\x12&\n"METRIC_KIND_RESOLVER_REQUEST_COUNT\x10\x07\x12 \n\x1cMETRIC_KIND_RESOLVER_LATENCY\x10\x08\x12&\n"METRIC_KIND_RESOLVER_SUCCESS_RATIO\x10\t\x12\x1b\n\x17METRIC_KIND_QUERY_COUNT\x10\n\x12\x1d\n\x19METRIC_KIND_QUERY_LATENCY\x10\x0b\x12#\n\x1fMETRIC_KIND_QUERY_SUCCESS_RATIO\x10\x0c\x12!\n\x1dMETRIC_KIND_BILLING_INFERENCE\x10\r\x12\x1c\n\x18METRIC_KIND_BILLING_CRON\x10\x0e\x12!\n\x1dMETRIC_KIND_BILLING_MIGRATION\x10\x0f\x12\x1a\n\x16METRIC_KIND_CRON_COUNT\x10\x10\x12\x1c\n\x18METRIC_KIND_CRON_LATENCY\x10\x11\x12)\n%METRIC_KIND_STREAM_MESSAGES_PROCESSED\x10\x12\x12&\n"METRIC_KIND_STREAM_MESSAGE_LATENCY\x10\x13\x12(\n$METRIC_KIND_STREAM_WINDOWS_PROCESSED\x10\x14\x12%\n!METRIC_KIND_STREAM_WINDOW_LATENCY\x10\x15\x12&\n"METRIC_KIND_ONLINE_STORE_KEY_COUNT\x10\x16\x12.\n*METRIC_KIND_ONLINE_STORE_EXPIRED_KEY_COUNT\x10\x17\x12\x30\n,METRIC_KIND_ONLINE_STORE_REQUESTS_PER_SECOND\x10\x18\x12\'\n#METRIC_KIND_CPU_UTILIZATION_PERCENT\x10\x19\x12\x1d\n\x19METRIC_KIND_REPLICA_COUNT\x10\x1a\x12(\n$METRIC_KIND_ONLINE_STORE_USED_MEMORY\x10\x1b\x12)\n%METRIC_KIND_ONLINE_STORE_TOTAL_MEMORY\x10\x1c\x12"\n\x1eMETRIC_KIND_MEMORY_USAGE_BYTES\x10\x1d\x12,\n(METRIC_KIND_TOTAL_MEMORY_AVAILABLE_BYTES\x10\x1e\x12"\n\x1eMETRIC_KIND_NETWORK_READ_BYTES\x10\x1f\x12#\n\x1fMETRIC_KIND_NETWORK_WRITE_BYTES\x10 \x12\x1f\n\x1bMETRIC_KIND_DISK_READ_BYTES\x10!\x12 \n\x1cMETRIC_KIND_DISK_WRITE_BYTES\x10"\x12\x1a\n\x16METRIC_KIND_STREAM_LAG\x10#\x12\x15\n\x11METRIC_KIND_USAGE\x10$*\xa3\x04\n\nFilterKind\x12\x1b\n\x17\x46ILTER_KIND_UNSPECIFIED\x10\x00\x12\x1e\n\x1a\x46ILTER_KIND_FEATURE_STATUS\x10\x01\x12\x1c\n\x18\x46ILTER_KIND_FEATURE_NAME\x10\x02\x12\x1b\n\x17\x46ILTER_KIND_FEATURE_TAG\x10\x03\x12\x1f\n\x1b\x46ILTER_KIND_RESOLVER_STATUS\x10\x04\x12\x1d\n\x19\x46ILTER_KIND_RESOLVER_NAME\x10\x05\x12\x1c\n\x18\x46ILTER_KIND_RESOLVER_TAG\x10\x06\x12\x1b\n\x17\x46ILTER_KIND_CRON_STATUS\x10\x07\x12 \n\x1c\x46ILTER_KIND_MIGRATION_STATUS\x10\x08\x12\x1e\n\x1a\x46ILTER_KIND_ONLINE_OFFLINE\x10\t\x12\x19\n\x15\x46ILTER_KIND_CACHE_HIT\x10\n\x12\x1c\n\x18\x46ILTER_KIND_OPERATION_ID\x10\x0b\x12\x1a\n\x16\x46ILTER_KIND_QUERY_NAME\x10\x0c\x12\x1c\n\x18\x46ILTER_KIND_QUERY_STATUS\x10\r\x12\x17\n\x13\x46ILTER_KIND_IS_NULL\x10\x0e\x12\x1a\n\x16\x46ILTER_KIND_USAGE_KIND\x10\x0f\x12\x1e\n\x1a\x46ILTER_KIND_RESOURCE_GROUP\x10\x10\x12\x18\n\x14\x46ILTER_KIND_POD_NAME\x10\x11*~\n\x0e\x43omparatorKind\x12\x1f\n\x1b\x43OMPARATOR_KIND_UNSPECIFIED\x10\x00\x12\x16\n\x12\x43OMPARATOR_KIND_EQ\x10\x01\x12\x17\n\x13\x43OMPARATOR_KIND_NEQ\x10\x02\x12\x1a\n\x16\x43OMPARATOR_KIND_ONE_OF\x10\x03*\xec\x03\n\x12WindowFunctionKind\x12$\n WINDOW_FUNCTION_KIND_UNSPECIFIED\x10\x00\x12\x1e\n\x1aWINDOW_FUNCTION_KIND_COUNT\x10\x01\x12\x1d\n\x19WINDOW_FUNCTION_KIND_MEAN\x10\x02\x12\x1c\n\x18WINDOW_FUNCTION_KIND_SUM\x10\x03\x12\x1c\n\x18WINDOW_FUNCTION_KIND_MIN\x10\x04\x12\x1c\n\x18WINDOW_FUNCTION_KIND_MAX\x10\x05\x12&\n"WINDOW_FUNCTION_KIND_PERCENTILE_99\x10\x06\x12&\n"WINDOW_FUNCTION_KIND_PERCENTILE_95\x10\x07\x12&\n"WINDOW_FUNCTION_KIND_PERCENTILE_75\x10\x08\x12&\n"WINDOW_FUNCTION_KIND_PERCENTILE_50\x10\t\x12&\n"WINDOW_FUNCTION_KIND_PERCENTILE_25\x10\n\x12%\n!WINDOW_FUNCTION_KIND_PERCENTILE_5\x10\x0b\x12(\n$WINDOW_FUNCTION_KIND_ALL_PERCENTILES\x10\x0c*\xe7\x03\n\x0bGroupByKind\x12\x1d\n\x19GROUP_BY_KIND_UNSPECIFIED\x10\x00\x12 \n\x1cGROUP_BY_KIND_FEATURE_STATUS\x10\x01\x12\x1e\n\x1aGROUP_BY_KIND_FEATURE_NAME\x10\x02\x12\x19\n\x15GROUP_BY_KIND_IS_NULL\x10\x03\x12!\n\x1dGROUP_BY_KIND_RESOLVER_STATUS\x10\x04\x12\x1f\n\x1bGROUP_BY_KIND_RESOLVER_NAME\x10\x05\x12\x1e\n\x1aGROUP_BY_KIND_QUERY_STATUS\x10\x06\x12\x1c\n\x18GROUP_BY_KIND_QUERY_NAME\x10\x07\x12 \n\x1cGROUP_BY_KIND_ONLINE_OFFLINE\x10\x08\x12\x1b\n\x17GROUP_BY_KIND_CACHE_HIT\x10\t\x12\x1c\n\x18GROUP_BY_KIND_USAGE_KIND\x10\n\x12 \n\x1cGROUP_BY_KIND_RESOURCE_GROUP\x10\x0b\x12\x1f\n\x1bGROUP_BY_KIND_DEPLOYMENT_ID\x10\x0c\x12\x1e\n\x1aGROUP_BY_KIND_OPERATION_ID\x10\r\x12\x1a\n\x16GROUP_BY_KIND_POD_NAME\x10\x0e*\xe4\x02\n\x11MetricFormulaKind\x12#\n\x1fMETRIC_FORMULA_KIND_UNSPECIFIED\x10\x00\x12\x1b\n\x17METRIC_FORMULA_KIND_SUM\x10\x01\x12#\n\x1fMETRIC_FORMULA_KIND_TOTAL_RATIO\x10\x02\x12\x1d\n\x19METRIC_FORMULA_KIND_RATIO\x10\x03\x12\x1f\n\x1bMETRIC_FORMULA_KIND_PRODUCT\x10\x04\x12\x1b\n\x17METRIC_FORMULA_KIND_ABS\x10\x05\x12\x1f\n\x1bMETRIC_FORMULA_KIND_KS_STAT\x10\x06\x12\x1f\n\x1bMETRIC_FORMULA_KIND_KS_TEST\x10\x07\x12$\n METRIC_FORMULA_KIND_KS_THRESHOLD\x10\x08\x12#\n\x1fMETRIC_FORMULA_KIND_TIME_OFFSET\x10\t*\xb8\x01\n\x11\x41lertSeverityKind\x12#\n\x1f\x41LERT_SEVERITY_KIND_UNSPECIFIED\x10\x00\x12 \n\x1c\x41LERT_SEVERITY_KIND_CRITICAL\x10\x01\x12\x1d\n\x19\x41LERT_SEVERITY_KIND_ERROR\x10\x02\x12\x1f\n\x1b\x41LERT_SEVERITY_KIND_WARNING\x10\x03\x12\x1c\n\x18\x41LERT_SEVERITY_KIND_INFO\x10\x04*\xdc\x01\n\rThresholdKind\x12\x1e\n\x1aTHRESHOLD_KIND_UNSPECIFIED\x10\x00\x12\x18\n\x14THRESHOLD_KIND_ABOVE\x10\x01\x12\x18\n\x14THRESHOLD_KIND_BELOW\x10\x02\x12 \n\x1cTHRESHOLD_KIND_GREATER_EQUAL\x10\x03\x12\x1d\n\x19THRESHOLD_KIND_LESS_EQUAL\x10\x04\x12\x18\n\x14THRESHOLD_KIND_EQUAL\x10\x05\x12\x1c\n\x18THRESHOLD_KIND_NOT_EQUAL\x10\x06*\xc7\x01\n\rChartLinkKind\x12\x1f\n\x1b\x43HART_LINK_KIND_UNSPECIFIED\x10\x00\x12\x1c\n\x18\x43HART_LINK_KIND_RESOLVER\x10\x01\x12\x1b\n\x17\x43HART_LINK_KIND_FEATURE\x10\x02\x12\x19\n\x15\x43HART_LINK_KIND_QUERY\x10\x03\x12\x1a\n\x16\x43HART_LINK_KIND_MANUAL\x10\x04\x12#\n\x1f\x43HART_LINK_KIND_SCHEDULED_QUERY\x10\x05\x42\x8e\x01\n\x16\x63om.chalk.artifacts.v1B\nChartProtoP\x01\xa2\x02\x03\x43\x41X\xaa\x02\x12\x43halk.Artifacts.V1\xca\x02\x12\x43halk\\Artifacts\\V1\xe2\x02\x1e\x43halk\\Artifacts\\V1\\GPBMetadata\xea\x02\x14\x43halk::Artifacts::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.artifacts.v1.chart_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\026com.chalk.artifacts.v1B\nChartProtoP\001\242\002\003CAX\252\002\022Chalk.Artifacts.V1\312\002\022Chalk\\Artifacts\\V1\342\002\036Chalk\\Artifacts\\V1\\GPBMetadata\352\002\024Chalk::Artifacts::V1"
    _globals["_METRICKIND"]._serialized_start = 1891
    _globals["_METRICKIND"]._serialized_end = 3238
    _globals["_FILTERKIND"]._serialized_start = 3241
    _globals["_FILTERKIND"]._serialized_end = 3788
    _globals["_COMPARATORKIND"]._serialized_start = 3790
    _globals["_COMPARATORKIND"]._serialized_end = 3916
    _globals["_WINDOWFUNCTIONKIND"]._serialized_start = 3919
    _globals["_WINDOWFUNCTIONKIND"]._serialized_end = 4411
    _globals["_GROUPBYKIND"]._serialized_start = 4414
    _globals["_GROUPBYKIND"]._serialized_end = 4901
    _globals["_METRICFORMULAKIND"]._serialized_start = 4904
    _globals["_METRICFORMULAKIND"]._serialized_end = 5260
    _globals["_ALERTSEVERITYKIND"]._serialized_start = 5263
    _globals["_ALERTSEVERITYKIND"]._serialized_end = 5447
    _globals["_THRESHOLDKIND"]._serialized_start = 5450
    _globals["_THRESHOLDKIND"]._serialized_end = 5670
    _globals["_CHARTLINKKIND"]._serialized_start = 5673
    _globals["_CHARTLINKKIND"]._serialized_end = 5872
    _globals["_ALERTTRIGGER"]._serialized_start = 55
    _globals["_ALERTTRIGGER"]._serialized_end = 445
    _globals["_DATASETFEATUREOPERAND"]._serialized_start = 447
    _globals["_DATASETFEATUREOPERAND"]._serialized_end = 522
    _globals["_METRICFORMULA"]._serialized_start = 525
    _globals["_METRICFORMULA"]._serialized_end = 872
    _globals["_METRICFILTER"]._serialized_start = 875
    _globals["_METRICFILTER"]._serialized_end = 1031
    _globals["_METRICCONFIGSERIES"]._serialized_start = 1034
    _globals["_METRICCONFIGSERIES"]._serialized_end = 1345
    _globals["_METRICCONFIG"]._serialized_start = 1348
    _globals["_METRICCONFIG"]._serialized_end = 1647
    _globals["_CHART"]._serialized_start = 1650
    _globals["_CHART"]._serialized_end = 1888
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import asyncio
import contextvars
import functools
import os
import threading
import weakref
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Any, Callable, TypeVar

from typing_extensions import Final, ParamSpec

DESIRED_CPU_PARALLELISM: int = int(os.getenv("OMP_NUM_THREADS", max(8, os.cpu_count() or 8)))
"""
The desired cpu parallelism. os.cpu_count() should never be used on k8s, as it returns the total
number of cores on the node rather than the requested cpu count from the pod manifest.
Rarely would os.cpu_count() return None, but falling back to 8 in that case.

We set a minimum bound of 8 for DESIRED_CPU_PARALLELISM because arrow can deadlock if initialized with too few threads.
"""


class MultiSemaphore:
    """Semaphore-like class that takes a value for both acquire and release"""

    def __init__(self, value: int = 1, /) -> None:
        super().__init__()
        self._value = value
        self.initial_value: Final = value
        self._cond = threading.Condition()

    def get_value(self):
        return self._value

    def acquire(self, val: int = 1, /, *, block: bool = True, timeout: float | None = None):
        if val <= 0:
            raise ValueError(f"Value ({val}) is not positive")
        if val > self.initial_value:
            raise ValueError(f"Value ({val}) is greater than the initial value ({self.initial_value})")
        if timeout is not None and timeout < 0:
            raise ValueError(f"Timeout ({timeout}) is negative, which is not supported")
        if not block:
            if timeout:
                raise ValueError("If `block` is False, then the timeout must not be specified (or be 0)")
            timeout = 0
        with self._cond:
            self._cond.wait_for(lambda: self._value - val >= 0, timeout)
            if self._value - val >= 0:
                self._value -= val
                return True
            return False

    def release(self, val: int):
        if val <= 0:
            raise ValueError(f"Value ({val}) is not positive")
        with self._cond:
            if self._value + val > self.initial_value:
                raise ValueError(f"Value ({val}) would put the semaphore above the initial value.")
            self._value += val
            self._cond.notify()


__all__ = ["ChalkThreadPoolExecutor"]

P = ParamSpec("P")
T = TypeVar("T")


class _ThreadlocalData(threading.local):
    parent_thread_pools: weakref.WeakSet[  # pyright: ignore[reportUninitializedInstanceVariable]
        ChalkThreadPoolExecutor
    ]


_threadlocal_data = _ThreadlocalData()


class ChalkThreadPoolExecutor(ThreadPoolExecutor):
    """Implementation of :class:`ThreadPoolExecutor` that adds trace context propagation,
    log context propagation, an asyncio hook, and thread pool checks."""

    def __init__(
        self,
        max_workers: int | None = None,
        thread_name_prefix: str = "",
        initializer: Callable[..., object] | None = None,
        initargs: tuple[Any, ...] = (),
    ) -> None:
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)

    if __debug__:

        def submit(  # pyright: ignore[reportRedeclaration]
            self, fn: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
        ) -> Future[T]:
            """Submit a new task to the thread pool."""
            if not hasattr(_threadlocal_data, "parent_thread_pools"):
                _threadlocal_data.parent_thread_pools = weakref.WeakSet()

            if self in _threadlocal_data.parent_thread_pools:
                raise RuntimeError(
                    (
                        "Attempted to submit a task to a thread pool from a thread spawned by the same thread pool. "
                        "This is a logic error which can lead to deadlocks."
                    )
                )
            # Copy the thread pool references and include `self` this time
            new_parent_thread_pools = weakref.WeakSet((*_threadlocal_data.parent_thread_pools, self))
            current_context = contextvars.copy_context()

            @functools.wraps(fn)
            def wrapped_fn(*args: P.args, **kwargs: P.kwargs):
                _threadlocal_data.parent_thread_pools = new_parent_thread_pools
                # Setting _merge=False to override the existing context,
                # in case if there is context left behind in the current thread from a previous
                # task
                return fn(*args, **kwargs)

            return super().submit(current_context.run, wrapped_fn, *args, **kwargs)

    else:
        # If not debug mode, skipping the reenetrency checking
        def submit(self, fn: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs) -> Future[T]:
            """Submit a new task to the thread pool."""
            current_context = contextvars.copy_context()
            return super().submit(current_context.run, fn, *args, **kwargs)

    async def submit_async(self, fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        return await asyncio.get_running_loop().run_in_executor(self, functools.partial(fn, *args, **kwargs))


DEFAULT_IO_EXECUTOR = ChalkThreadPoolExecutor(
    max_workers=int(os.getenv("CHALK_IO_EXECUTOR_MAX_WORKERS", "32")),
    thread_name_prefix="chalk-io-",
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/auth/v1/displayagent.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import featurepermission_pb2 as chalk_dot_auth_dot_v1_dot_featurepermission__pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n chalk/auth/v1/displayagent.proto\x12\rchalk.auth.v1\x1a%chalk/auth/v1/featurepermission.proto\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a\x1fgoogle/protobuf/timestamp.proto"b\n\x11\x44isplayPermission\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x39\n\npermission\x18\x02 \x01(\x0e\x32\x19.chalk.auth.v1.PermissionR\npermission"\xf7\x01\n\x1d\x44isplayEnvironmentPermissions\x12%\n\x0e\x65nvironment_id\x18\x01 \x01(\tR\renvironmentId\x12)\n\x10\x65nvironment_name\x18\x02 \x01(\tR\x0f\x65nvironmentName\x12\x1d\n\nproject_id\x18\x03 \x01(\tR\tprojectId\x12!\n\x0cproject_name\x18\x04 \x01(\tR\x0bprojectName\x12\x42\n\x0bpermissions\x18\x05 \x03(\x0b\x32 .chalk.auth.v1.DisplayPermissionR\x0bpermissions"\x88\x03\n\x10\x44isplayUserAgent\x12\x1f\n\tclient_id\x18\x01 \x01(\tB\x02\x18\x01R\x08\x63lientId\x12\x17\n\x07user_id\x18\x02 \x01(\tR\x06userId\x12\x17\n\x07team_id\x18\x03 \x01(\tR\x06teamId\x12j\n\x1apermissions_by_environment\x18\x04 \x03(\x0b\x32,.chalk.auth.v1.DisplayEnvironmentPermissionsR\x18permissionsByEnvironment\x12"\n\x0cimpersonated\x18\x05 \x01(\x08R\x0cimpersonated\x12\x12\n\x04name\x18\x06 \x01(\tR\x04name\x12\x14\n\x05\x65mail\x18\x07 \x01(\tR\x05\x65mail\x12\x1b\n\tteam_name\x18\x08 \x01(\tR\x08teamName\x12J\n\x0f\x63ustomer_claims\x18\t \x03(\x0b\x32!.chalk.auth.v1.DisplayCustomClaimR\x0e\x63ustomerClaims">\n\x12\x44isplayCustomClaim\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x16\n\x06values\x18\x02 \x03(\tR\x06values"\xc4\x04\n\x18\x44isplayServiceTokenAgent\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x1b\n\tclient_id\x18\x02 \x01(\tR\x08\x63lientId\x12\x17\n\x07team_id\x18\x03 \x01(\tR\x06teamId\x12\x1d\n\nproject_id\x18\x04 \x01(\tR\tprojectId\x12%\n\x0e\x65nvironment_id\x18\x05 \x01(\tR\renvironmentId\x12\x42\n\x0bpermissions\x18\x06 \x03(\x0b\x32 .chalk.auth.v1.DisplayPermissionR\x0bpermissions\x12\x12\n\x04name\x18\x07 \x01(\tR\x04name\x12\x1b\n\tteam_name\x18\x08 \x01(\tR\x08teamName\x12!\n\x0cproject_name\x18\t \x01(\tR\x0bprojectName\x12)\n\x10\x65nvironment_name\x18\n \x01(\tR\x0f\x65nvironmentName\x12J\n\x0f\x63ustomer_claims\x18\x0b \x03(\x0b\x32!.chalk.auth.v1.DisplayCustomClaimR\x0e\x63ustomerClaims\x12R\n\x13\x66\x65\x61ture_permissions\x18\x0c \x01(\x0b\x32!.chalk.auth.v1.FeaturePermissionsR\x12\x66\x65\x61turePermissions\x12\x39\n\ncreated_at\x18\r \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt"\x92\x02\n\x12\x44isplayEngineAgent\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x17\n\x07team_id\x18\x02 \x01(\tR\x06teamId\x12\x1d\n\nproject_id\x18\x03 \x01(\tR\tprojectId\x12%\n\x0e\x65nvironment_id\x18\x04 \x01(\tR\renvironmentId\x12"\n\x0cimpersonated\x18\x05 \x01(\x08R\x0cimpersonated\x12\x1b\n\tteam_name\x18\x06 \x01(\tR\x08teamName\x12!\n\x0cproject_name\x18\x07 \x01(\tR\x0bprojectName\x12)\n\x10\x65nvironment_name\x18\x08 \x01(\tR\x0f\x65nvironmentName"\x14\n\x12\x44isplayTenantAgent"\x1d\n\x1b\x44isplayMetadataServiceAgent"\xa8\x03\n\x0c\x44isplayAgent\x12@\n\nuser_agent\x18\x01 \x01(\x0b\x32\x1f.chalk.auth.v1.DisplayUserAgentH\x00R\tuserAgent\x12Y\n\x13service_token_agent\x18\x02 \x01(\x0b\x32\'.chalk.auth.v1.DisplayServiceTokenAgentH\x00R\x11serviceTokenAgent\x12\x46\n\x0c\x65ngine_agent\x18\x03 \x01(\x0b\x32!.chalk.auth.v1.DisplayEngineAgentH\x00R\x0b\x65ngineAgent\x12\x46\n\x0ctenant_agent\x18\x04 \x01(\x0b\x32!.chalk.auth.v1.DisplayTenantAgentH\x00R\x0btenantAgent\x12\x62\n\x16metadata_service_agent\x18\x05 \x01(\x0b\x32*.chalk.auth.v1.DisplayMetadataServiceAgentH\x00R\x14metadataServiceAgentB\x07\n\x05\x61gentB\x8c\x01\n\x11\x63om.chalk.auth.v1B\x11\x44isplayagentProtoP\x01Z\x0e\x61uth/v1;authv1\xa2\x02\x03\x43\x41X\xaa\x02\rChalk.Auth.V1\xca\x02\rChalk\\Auth\\V1\xe2\x02\x19\x43halk\\Auth\\V1\\GPBMetadata\xea\x02\x0f\x43halk::Auth::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.auth.v1.displayagent_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\021com.chalk.auth.v1B\021DisplayagentProtoP\001Z\016auth/v1;authv1\242\002\003CAX\252\002\rChalk.Auth.V1\312\002\rChalk\\Auth\\V1\342\002\031Chalk\\Auth\\V1\\GPBMetadata\352\002\017Chalk::Auth::V1"
    _globals["_DISPLAYUSERAGENT"].fields_by_name["client_id"]._options = None
    _globals["_DISPLAYUSERAGENT"].fields_by_name["client_id"]._serialized_options = b"\030\001"
    _globals["_DISPLAYPERMISSION"]._serialized_start = 156
    _globals["_DISPLAYPERMISSION"]._serialized_end = 254
    _globals["_DISPLAYENVIRONMENTPERMISSIONS"]._serialized_start = 257
    _globals["_DISPLAYENVIRONMENTPERMISSIONS"]._serialized_end = 504
    _globals["_DISPLAYUSERAGENT"]._serialized_start = 507
    _globals["_DISPLAYUSERAGENT"]._serialized_end = 899
    _globals["_DISPLAYCUSTOMCLAIM"]._serialized_start = 901
    _globals["_DISPLAYCUSTOMCLAIM"]._serialized_end = 963
    _globals["_DISPLAYSERVICETOKENAGENT"]._serialized_start = 966
    _globals["_DISPLAYSERVICETOKENAGENT"]._serialized_end = 1546
    _globals["_DISPLAYENGINEAGENT"]._serialized_start = 1549
    _globals["_DISPLAYENGINEAGENT"]._serialized_end = 1823
    _globals["_DISPLAYTENANTAGENT"]._serialized_start = 1825
    _globals["_DISPLAYTENANTAGENT"]._serialized_end = 1845
    _globals["_DISPLAYMETADATASERVICEAGENT"]._serialized_start = 1847
    _globals["_DISPLAYMETADATASERVICEAGENT"]._serialized_end = 1876
    _globals["_DISPLAYAGENT"]._serialized_start = 1879
    _globals["_DISPLAYAGENT"]._serialized_end = 2303
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import TYPE_CHECKING, List, Mapping, Optional, Sequence, Type, Union

import pyarrow as pa

if TYPE_CHECKING:
    from pydantic import BaseModel
else:
    try:
        from pydantic.v1 import BaseModel
    except ImportError:
        from pydantic import BaseModel

from chalk.features.underscore import Underscore, UnderscoreFunction
from chalk.utils.pydanticutil.pydantic_compat import get_pydantic_output_structure


def message(role: str | Underscore, content: str | Underscore):
    return UnderscoreFunction(
        "struct_pack",
        ["role", "content"],
        role,
        content,
    )


def run_prompt(name: str):
    """
    Runs a named prompt. Configure named prompts in the UI.

    Parameters
    ----------
    name
        The name of the prompt to run.

    Examples
    --------
    >>> import chalk.prompts as P
    >>> from chalk.features import features
    >>> @features
    ... class User:
    ...    id: str
    ...    description: P.PromptResponse = P.run_prompt("describe_user")
    """
    return UnderscoreFunction("run_prompt", prompt_name=name)


def completion(
    model: str,
    messages: Sequence[Underscore],
    *,
    timeout_seconds: Optional[float] = None,
    output_structure: Optional[Union[Type[BaseModel], str]] = None,
    temperature: Optional[float | Underscore] = None,
    top_p: Optional[float | Underscore] = None,
    max_completion_tokens: Optional[int | Underscore] = None,
    max_tokens: Optional[int | Underscore] = None,
    stop: Optional[Sequence[str]] = None,
    presence_penalty: Optional[float | Underscore] = None,
    frequency_penalty: Optional[float | Underscore] = None,
    logit_bias: Optional[Mapping[int, float]] = None,
    seed: Optional[int | Underscore] = None,
    user: Optional[str | Underscore] = None,
    model_provider: Optional[str | Underscore] = None,
    base_url: Optional[str | Underscore] = None,
    api_key: Optional[str | Underscore] = None,
    num_retries: Optional[int | Underscore] = None,
):
    """
    Generate LLM model completions from a list of messages.

    Parameters
    ----------
    model
        The name of the model, e.g. "gpt-4o".
    messages
        The list of messages of the type P.Message. Each message in the array contains the following properties: role and content.
        The role of the message's author. Roles can be: system, user, or assistant.
        The contents of the message. It is required for all messages.
    timeout_seconds
        The timeout in seconds for completion requests
    output_structure
        The object specifying the format that the model must output. Accepts a Pydantic model or a JSON schema string (see https://docs.pydantic.dev/1.10/usage/schema/).
    temperature
        The sampling temperature to be used, between 0 and 2 inclusive. Higher values like 0.8 produce more random outputs, while lower values like 0.2 make outputs more focused and deterministic.
        Note: This parameter is between 0 and 1 (inclusive) for Anthropic models.
    top_p
        The alternative to sampling with temperature. It instructs the model to consider the results of the tokens with top_p probability. For example, 0.1 means only the tokens comprising the top 10% probability mass are considered.
    max_completion_tokens
        The upper bound for the number of tokens that can be generated for a completion, including visible output tokens and reasoning tokens.
    max_tokens
        The maximum number of tokens to generate in the chat completion.
    stop
        Custom text sequences that will cause the model to stop generating.
    presence_penalty
        It is used to penalize new tokens based on their existence in the text so far.
    frequency_penalty
        It is used to penalize new tokens based on their frequency in the text so far.
    logit_bias
        Used to modify the probability of specific tokens appearing in the completion.
    seed
        If specified, the system will make a best effort to sample deterministically, such that repeated requests with the same seed and parameters should return the same result. Determinism is not guaranteed.
    user
        The unique identifier representing your end-user. This parameter is specific to OpenAI and can help to monitor and detect abuse.
    model_provider
        The model provider.
    base_url
        The URL of the API endpoint where requests are sent.
    api_key
        The API key to use for the completion.
    num_retries
        The number of times to retry the API call if an APIError, TimeoutError or ServiceUnavailableError occurs.

    Examples
    --------
    >>> import chalk.prompts as P
    >>> import chalk.functions as F
    >>> from chalk.features import features, DataFrame, _
    >>> @features
    ... class User:
    ...    id: str
    ...    description: str
    ...    estimated_age_response: P.PromptResponse = P.completion(
    ...        model="gpt-4o",
    ...        messages=[
    ...            P.message("system", F.jinja("Estimate the age of the user based on the description: {{User.description}}")),
    ...        ],
    ...        max_tokens=100+2*F.length(_.description)
    ...    )
    ...    estimated_age: float = F.json_value(_.estimated_age_response.response, "$.age")
    """
    if isinstance(messages, str) or isinstance(messages, Underscore):
        raise ValueError("Messages should be a list of P.message objects, not a single object.")
    messages_parsed = UnderscoreFunction(
        "array_constructor",
        *messages,
    )
    if output_structure is None:
        output_structure_json = None
    elif isinstance(output_structure, str):
        output_structure_json = output_structure
    else:
        output_structure_json = get_pydantic_output_structure(output_structure)

    return UnderscoreFunction(
        "completion",
        model=model,
        messages=messages_parsed,
        timeout_seconds=timeout_seconds,
        output_structure=output_structure_json,
        temperature=temperature,
        top_p=top_p,
        max_completion_tokens=max_completion_tokens,
        max_tokens=max_tokens,
        stop=stop,
        presence_penalty=presence_penalty,
        frequency_penalty=frequency_penalty,
        logit_bias=pa.scalar(list(logit_bias.items()), type=pa.map_(pa.int64(), pa.float64()))
        if logit_bias is not None
        else None,
        seed=seed,
        user=user,
        num_retries=num_retries,
        model_provider=model_provider,
        base_url=base_url,
        api_key=api_key,
    )


class Message(BaseModel):
    role: str
    content: str


class Prompt(BaseModel):
    """Chalk Prompts are bundles of a model, a list of messages, and parameters."""

    model: str
    """The name of the model, e.g. "gpt-4o"."""
    messages: List[Message]
    """The list of messages of the type P.Message. Each message in the array contains the following properties: role and content.
    The role of the message's author. Roles can be: system, user, or assistant.
    The contents of the message. It is required for all messages."""
    timeout_seconds: Optional[float] = None
    """The timeout in seconds for completion requests."""
    output_structure: Optional[str] = None
    """The object specifying the format that the model must output. Accepts a Pydantic model or a JSON schema string (see https://docs.pydantic.dev/1.10/usage/schema/)."""
    temperature: Optional[float] = None
    """The sampling temperature to be used, between 0 and 2 inclusive. Higher values like 0.8 produce more random outputs, while lower values like 0.2 make outputs more focused and deterministic.
    Note: This parameter is between 0 and 1 (inclusive) for Anthropic models."""
    top_p: Optional[float] = None
    """The alternative to sampling with temperature. It instructs the model to consider the results of the tokens with top_p probability. For example, 0.1 means only the tokens comprising the top 10% probability mass are considered."""
    max_completion_tokens: Optional[int] = None
    """The upper bound for the number of tokens that can be generated for a completion, including visible output tokens and reasoning tokens."""
    max_tokens: Optional[int] = None
    """The maximum number of tokens to generate in the chat completion."""
    stop: Optional[List[str]] = None
    """Custom text sequences that will cause the model to stop generating."""
    presence_penalty: Optional[float] = None
    """It is used to penalize new tokens based on their existence in the text so far."""
    frequency_penalty: Optional[float] = None
    """It is used to penalize new tokens based on their frequency in the text so far."""
    logit_bias: Optional[Mapping[int, float]] = None
    """Used to modify the probability of specific tokens appearing in the completion."""
    seed: Optional[int] = None
    """If specified, the system will make a best effort to sample deterministically, such that repeated requests with the same seed and parameters should return the same result. Determinism is not guaranteed."""
    user: Optional[str] = None
    """The unique identifier representing your end-user. This parameter is specific to OpenAI and can help to monitor and detect abuse."""
    model_provider: Optional[str] = None
    """The model provider."""
    base_url: Optional[str] = None
    """The URL of the API endpoint where requests are sent."""
    num_retries: Optional[int] = None
    """The number of times to retry the API call if an APIError, TimeoutError or ServiceUnavailableError occurs."""


class Usage(BaseModel):
    """Usage statistics for the response."""

    input_tokens: int
    """Number of tokens in the request."""
    output_tokens: int
    """Number of tokens in the response."""
    total_tokens: int
    """Total number of tokens used, equal to input_tokens + output_tokens."""


class RuntimeStats(BaseModel):
    """Runtime statistics for the response."""

    total_latency: float
    """Total time in seconds to generate the response, including any retries."""
    last_try_latency: Optional[float]
    """Time in seconds to generate the response in the last successful try."""
    total_retries: int
    """Total number of retries."""
    rate_limit_retries: int
    """Number of retries due to rate limiting."""


class PromptResponse(BaseModel):
    """Response from the model."""

    response: Optional[str]
    """Response from the model. Raw string if no output structure specified, json encoded string otherwise. None if the response was not received or incorrectly formatted."""
    prompt: Prompt
    """Prompt used to generate the response."""
    usage: Usage
    """Usage statistics for the response."""
    runtime_stats: RuntimeStats
    """Runtime statistics for the response."""



from __future__ import annotations

import asyncio
import contextlib
import inspect
import json
import logging
import os
import os.path
import traceback
import warnings
from contextvars import ContextVar
from dataclasses import dataclass
from enum import Enum
from os import PathLike
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import pyarrow as pa

from chalk.clogging import chalk_logger
from chalk.features import Feature, FeatureConverter, Features, FeatureWrapper, unwrap_feature
from chalk.integrations.named import load_integration_variable
from chalk.sql._internal.chalk_query import ChalkQuery
from chalk.sql._internal.incremental import IncrementalSettings
from chalk.sql._internal.query_execution_parameters import (
    QueryExecutionParameters,
    query_execution_parameters_from_env_vars,
)
from chalk.sql._internal.string_chalk_query import StringChalkQuery
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.sql.protocols import BaseSQLSourceProtocol, ChalkQueryProtocol, StringChalkQueryProtocol, TableIngestProtocol
from chalk.utils.async_helpers import async_null_context, to_async_iterable
from chalk.utils.environment_parsing import env_var_bool
from chalk.utils.log_with_context import get_logger, get_logging_context
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import sqlalchemy.ext.asyncio
    from sqlalchemy.engine import URL, Connection, Dialect, Engine
    from sqlalchemy.orm import Session
    from sqlalchemy.sql import Select
    from sqlalchemy.sql.compiler import SQLCompiler
    from sqlalchemy.sql.ddl import CreateTable, DropTable
    from sqlalchemy.sql.elements import Label
    from sqlalchemy.sql.schema import Table
    from sqlalchemy.types import TypeEngine


TTableIngestMixIn = TypeVar("TTableIngestMixIn", bound="TableIngestMixIn")
CHALK_QUERY_LOGGING = env_var_bool("CHALK_QUERY_LOGGING")

_logger = get_logger(__name__)

_ENABLE_ADD_TO_SQL_SOURCE_REGISTRIES: ContextVar[bool] = ContextVar(
    "_ENABLE_ADD_TO_SQL_SOURCE_REGISTRIES", default=True
)


@dataclass
class TableIngestionPreferences:
    features: Type[Features]
    ignore_columns: Set[str]
    ignore_features: Set[str]
    require_columns: Set[str]
    require_features: Set[str]
    column_to_feature: Dict[str, str]
    cdc: Optional[Union[bool, IncrementalSettings]]


def _force_set_str(x: Optional[List[Any]]) -> Set[str]:
    return set() if x is None else set(map(str, x))


class UnsupportedEfficientExecutionError(ValueError):
    def __init__(self, msg: str, log_level: int) -> None:
        super().__init__(msg)
        self.log_level = log_level


def validate_dtypes_for_efficient_execution(stmt: Select, supported_types: Sequence[Type[TypeEngine]]):
    unsupported_columns = [
        x
        for x in stmt.column_descriptions
        # Using NOT IN rather than not isinstance() as subclasses may override the result processor,
        # which makes it not eligible for efficient execution
        if x["type"].__class__ not in supported_types
    ]
    if len(unsupported_columns) == 1 and unsupported_columns[0]["name"] == "*":
        _logger.warning(
            "Got '*' for select clause. Will try efficient execution and fallback if unsupported types are found."
        )
        return

    if len(unsupported_columns) > 0:
        unsupported_columns_and_dtypes = [(x["name"], x["type"].__class__.__name__) for x in unsupported_columns]
        supported_dtypes = [x.__name__ for x in supported_types]
        formatted_supported = ", ".join(supported_dtypes)
        formatted_unsupported = ", ".join([f"{name} ({dtype})" for (name, dtype) in unsupported_columns_and_dtypes])
        raise UnsupportedEfficientExecutionError(
            (
                "The SQL statement will be executed into SQLAlchemy objects, as the SQL query returns columns "
                "that cannot be loaded directly into a PyArrow table. For better performance, use only "
                f"the following types in your SQLAlchemy model: {formatted_supported}."
                f"The columns that contain unsupported types are: {formatted_unsupported}."
            ),
            log_level=logging.INFO,
        )


class TableIngestMixIn(TableIngestProtocol):
    ingested_tables: Dict[str, TableIngestionPreferences]

    def with_table(
        self: TTableIngestMixIn,
        *,
        name: str,
        features: Type[Union[Features, Any]],
        ignore_columns: Optional[List[str]] = None,
        ignore_features: Optional[List[Union[str, Any]]] = None,
        require_columns: Optional[List[str]] = None,
        require_features: Optional[List[Union[str, Any]]] = None,
        column_to_feature: Optional[Dict[str, Any]] = None,
        cdc: Optional[Union[bool, IncrementalSettings]] = None,
    ) -> TTableIngestMixIn:
        if name in self.ingested_tables:
            raise ValueError(f"The table {name} is ingested twice.")
        self.ingested_tables[name] = TableIngestionPreferences(
            features=features,
            ignore_columns=_force_set_str(ignore_columns),
            ignore_features=_force_set_str(ignore_features),
            require_columns=_force_set_str(require_columns),
            require_features=_force_set_str(require_features),
            column_to_feature={k: str(v) for k, v in (column_to_feature or {}).items()},
            cdc=cdc,
        )
        return self


class SQLSourceKind(str, Enum):
    bigquery = "bigquery"
    clickhouse = "clickhouse"
    cloudsql = "cloudsql"
    databricks = "databricks"
    athena = "athena"
    duckdb = "duckdb"
    dynamodb = "dynamodb"
    mysql = "mysql"
    postgres = "postgres"
    redshift = "redshift"
    snowflake = "snowflake"
    spanner = "spanner"
    sqlite = "sqlite"
    trino = "trino"

    @staticmethod
    def convert_sql_source_kind(kind: str) -> SQLSourceKind:
        try:
            return SQLSourceKind(kind)
        except ValueError:
            raise ValueError(f"Unsupported SQL source kind: {kind}")


class BaseSQLSource(BaseSQLSourceProtocol):
    registry: ClassVar[List["BaseSQLSource"]] = []

    kind: SQLSourceKind

    def __init__(
        self,
        name: Optional[str],
        engine_args: Optional[Dict[str, Any]],
        async_engine_args: Optional[Dict[str, Any]],
    ):
        super().__init__()
        try:
            import sqlalchemy
        except ImportError:
            raise missing_dependency_exception("chalkpy[sql]")
        del sqlalchemy  # unused

        self._incremental_settings = None
        self._resolver_and_sqlfile_to_sqlstring: Dict[Tuple[str, str], str] = {}
        if _ENABLE_ADD_TO_SQL_SOURCE_REGISTRIES.get():
            self.registry.append(self)
        self.name = name
        if engine_args is None:
            engine_args = {}
        if async_engine_args is None:
            async_engine_args = {}
        if self.name is not None:
            for k, v in self._load_env_engine_args(name=self.name, override=None).items():
                engine_args.setdefault(k, v)
                async_engine_args.setdefault(k, v)
        if getattr(self, "kind", None) != SQLSourceKind.trino:
            engine_args.setdefault("pool_pre_ping", env_var_bool("USE_CLIENT_POOL_PRE_PING"))
            async_engine_args.setdefault("pool_pre_ping", env_var_bool("USE_CLIENT_POOL_PRE_PING"))
        self.engine_args = engine_args
        self.async_engine_args = async_engine_args
        self._engine = None
        self._async_engine = None

    @property
    def _engine_args(self):
        """Backcompat support for private subclassing of BaseSQLSource"""
        return self.engine_args

    @_engine_args.setter
    def _engine_args(self, args: dict[str, Any]):
        """Backcompat support for private subclassing of BaseSQLSource"""
        self.engine_args = args

    @property
    def _async_engine_args(self):
        """Backcompat support for private subclassing of BaseSQLSource"""
        return self.async_engine_args

    @_async_engine_args.setter
    def _async_engine_args(self, args: dict[str, Any]):
        """Backcompat support for private subclassing of BaseSQLSource"""
        self.async_engine_args = args

    def get_sqlglot_dialect(self) -> Union[str, None]:
        """Returns the name of the SQL dialect (if it has one) for `sqlglot` to parse the SQL string.
        This allows for use of dialect-specific syntax while parsing and modifying queries."""
        return None

    def supports_inefficient_fallback(self) -> bool:
        return True

    def query_sql_file(
        self,
        path: Union[str, bytes, PathLike],
        fields: Optional[Mapping[str, Union[Feature, str, Any]]] = None,
        args: Optional[Mapping[str, object]] = None,
    ) -> StringChalkQueryProtocol:
        sql_string = None
        if isinstance(path, bytes):
            path = path.decode("utf-8")

        resolver_fqn = get_logging_context().get("labels", {}).get("resolver_fqn")
        if resolver_fqn is not None:
            sql_string = self._resolver_and_sqlfile_to_sqlstring.get((resolver_fqn, str(path)))
        uncached = sql_string is None

        if uncached:
            _logger.info(f"SQL query for resolver '{resolver_fqn}' from file '{str(path)}' is not cached")
            if os.path.isfile(path):
                with open(path) as f:
                    sql_string = f.read()
            else:
                frame = inspect.currentframe()
                assert frame is not None
                caller_frame = frame.f_back
                del frame
                assert caller_frame is not None
                while caller_frame.f_globals["__name__"].startswith("chalk.sql") and caller_frame.f_back is not None:
                    caller_frame = caller_frame.f_back
                dir_path = os.path.dirname(os.path.realpath(inspect.getframeinfo(caller_frame).filename))
                path = str(path)
                relative_path = os.path.join(dir_path, path)
                if os.path.isfile(relative_path):
                    with open(relative_path) as f:
                        sql_string = f.read()
        if sql_string is None:
            raise FileNotFoundError(f"No such file: '{str(path)}'")

        if uncached and resolver_fqn is not None:
            # Caching by the resolver fqn because the file path could be relative
            # to the file that the resolver is defined in
            self._resolver_and_sqlfile_to_sqlstring[(resolver_fqn, str(path))] = sql_string

        return self.query_string(
            query=sql_string,
            fields=fields,
            args=args,
        )

    def query_string(
        self,
        query: str,
        fields: Optional[Mapping[str, Union[Feature, Any]]] = None,
        args: Optional[Mapping[str, object]] = None,
    ) -> StringChalkQueryProtocol:
        fields = fields or {}
        fields = {f: unwrap_feature(v) if isinstance(v, FeatureWrapper) else v for (f, v) in fields.items()}
        return StringChalkQuery(source=self, query=query, fields=fields, params=args or {})

    def warm_up(self) -> None:
        from sqlalchemy import text

        with self.get_engine().connect() as cnx:
            cnx.execute(text("select 1"))

    def query(self, *entities: Any) -> ChalkQueryProtocol:
        targets: List[Label] = []
        features: Dict[str, Feature] = {}

        for e in entities:
            if isinstance(e, Features):
                _extract_features(e, e.__chalk_namespace__, targets, features)
            else:
                targets.append(e)

        return ChalkQuery(
            features=features,
            targets=targets,
            source=self,
        )

    def local_engine_url(self) -> URL:
        raise NotImplementedError

    def async_local_engine_url(self) -> URL:
        raise NotImplementedError

    def execute_query(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[Connection] = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> Iterable[pa.RecordBatch]:
        if query_execution_parameters is None:
            """
            Sometimes, in tests or in development or in user resolver code,
            we may not have the query_execution_parameters argument
            passed in. Thus, we need to hack the env vars in.
            """
            query_execution_parameters = query_execution_parameters_from_env_vars()

        efficient_execution_failed: bool = False
        efficient_execution_traceback: str | None = None
        attempt_efficient_execution = query_execution_parameters.attempt_efficient_execution

        if CHALK_QUERY_LOGGING:
            log_dialect = None
            try:
                log_dialect = self.get_sqlalchemy_dialect(paramstyle="pyformat")
            except Exception as e:
                _logger.warning(f"Could not determine sqlalchemy dialect of source of type {self.kind}", exc_info=e)
            _logger.info(
                f"Executing SQL query: {finalized_query.query.compile(compile_kwargs={'literal_binds': True}, dialect=log_dialect)}"
            )

        if attempt_efficient_execution:
            try:
                yield from self._execute_query_efficient(
                    finalized_query, columns_to_features, connection, query_execution_parameters
                )
                return
            except NotImplementedError:
                _logger.debug(
                    (
                        "The SQL statement will be executed into SQLAlchemy objects, as the database backend does "
                        "not support a more efficient execution mechanism."
                    )
                )
                pass
            except UnsupportedEfficientExecutionError as e:
                log_level = e.log_level
                _logger.log(log_level, str(e))
            except Exception:
                if (
                    query_execution_parameters.fallback_to_inefficient_execution
                    and self.supports_inefficient_fallback()
                ):
                    _logger.error(
                        (
                            f"Failed to efficiently execute query {finalized_query.query.compile(compile_kwargs={'literal_binds': True})} "
                            f"with parameters {finalized_query.params} due to an unexpected error. Falling back to inefficient execution."
                        ),
                        exc_info=True,
                    )
                    efficient_execution_failed = True
                    efficient_execution_traceback = traceback.format_exc()
                else:
                    raise
        try:
            batch = self._execute_query_inefficient(
                finalized_query, columns_to_features, connection, query_execution_parameters
            )
        except:
            raise
        else:
            if efficient_execution_failed:
                # This helps us collect into about the causes of errors in the 'efficient' code path.
                # i.e. are they code bugs we can fix, or actual errors e.g. in the database itself?
                _logger.info(
                    f"Failed to efficiently execute query, but inefficient code path succeeded. Original error was: {efficient_execution_traceback}"
                )
            if len(batch) > 0 or query_execution_parameters.yield_empty_batches:
                yield batch

    @contextlib.contextmanager
    def _create_temp_table(
        self,
        create_temp_table: CreateTable,
        temp_table: Table,
        drop_temp_table: DropTable,
        connection: sqlalchemy.engine.Connection,
        temp_value: pa.Table,
    ):
        chalk_logger.info(f"Creating temporary table {temp_table.name}.")
        connection.execute(create_temp_table)
        try:
            batch_size = insert_size_limit // temp_value.num_columns
            assert temp_value.num_columns <= insert_size_limit, "temp temp_value has too many columns to insert rowwise"
            for batch_start in range(0, len(temp_value), batch_size):
                batch_end = min(batch_start + batch_size, len(temp_value))
                connection.execute(
                    temp_table.insert(), temp_value.slice(batch_start, batch_end - batch_start).to_pylist()
                )
            yield
        finally:
            # "temp table", to snowflake, means that it belongs to the session. However, we keep using the same Snowflake session
            chalk_logger.info(f"Dropping temporary table {temp_table.name}.")
            connection.execute(drop_temp_table)

    @contextlib.asynccontextmanager
    async def _async_create_temp_table(
        self,
        create_temp_table: CreateTable,
        temp_table: Table,
        drop_temp_table: DropTable,
        connection: sqlalchemy.ext.asyncio.AsyncConnection,
        temp_value: pa.Table,
    ):
        chalk_logger.info(f"Creating temporary table {temp_table.name}.")
        await connection.execute(create_temp_table)
        try:
            batch_size = insert_size_limit // temp_value.num_columns
            assert temp_value.num_columns <= insert_size_limit, "temp temp_value has too many columns to insert rowwise"
            for batch_start in range(0, len(temp_value), batch_size):
                batch_end = min(batch_start + batch_size, len(temp_value))
                await connection.execute(
                    temp_table.insert(), temp_value.slice(batch_start, batch_end - batch_start).to_pylist()
                )
            yield
        finally:
            # "temp table", to snowflake, means that it belongs to the session. However, we keep using the same Snowflake session
            chalk_logger.info(f"Dropping temporary table {temp_table.name}.")
            await connection.execute(drop_temp_table)

    async def async_execute_query(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection] = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> AsyncIterable[pa.RecordBatch]:
        if query_execution_parameters is None:
            query_execution_parameters = query_execution_parameters_from_env_vars()
        attempt_efficient_execution = query_execution_parameters.attempt_efficient_execution
        efficient_execution_failed: bool = False
        efficient_execution_traceback: str | None = None
        if attempt_efficient_execution:
            try:
                async for res in self._async_execute_query_efficient(
                    finalized_query, columns_to_features, connection, query_execution_parameters
                ):
                    yield res
                return
            except NotImplementedError:
                try:
                    async for batch in to_async_iterable(
                        self._execute_query_efficient(
                            finalized_query,
                            columns_to_features,
                            None,  # async connection is not compatible with a non-async connection
                            query_execution_parameters,
                        )
                    ):
                        yield batch
                    return
                except NotImplementedError:
                    _logger.debug(
                        (
                            "The SQL statement will be executed into SQLAlchemy objects, as the database backend does "
                            "not support a more efficient execution mechanism."
                        )
                    )
                    pass
                except UnsupportedEfficientExecutionError as e:
                    log_level = e.log_level
                    _logger.log(log_level, str(e))
                except Exception:
                    if query_execution_parameters.fallback_to_inefficient_execution:
                        _logger.error(
                            (
                                f"Failed to efficiently execute query {finalized_query.query.compile(compile_kwargs={'literal_binds': True})} "
                                f"with parameters {finalized_query.params} due to an unexpected error. Falling back to inefficient execution."
                            ),
                            exc_info=True,
                        )
                        efficient_execution_failed = True
                        efficient_execution_traceback = traceback.format_exc()
                    else:
                        raise
            except UnsupportedEfficientExecutionError as e:
                log_level = e.log_level
                _logger.log(log_level, str(e))
            except Exception:
                if query_execution_parameters.fallback_to_inefficient_execution:
                    _logger.error(
                        (
                            f"Failed to efficiently execute query {finalized_query.query.compile(compile_kwargs={'literal_binds': True})} "
                            f"with parameters {finalized_query.params} due to an unexpected error. Falling back to inefficient execution."
                        ),
                        exc_info=True,
                    )
                    efficient_execution_failed = True
                    efficient_execution_traceback = traceback.format_exc()
                else:
                    raise
        try:
            batch = await self._async_execute_query_inefficient(
                finalized_query, columns_to_features, connection, query_execution_parameters
            )
        except:
            raise
        else:
            if efficient_execution_failed:
                # This helps us collect into about the causes of errors in the 'efficient' code path.
                # i.e. are they code bugs we can fix, or actual errors e.g. in the database itself?
                _logger.info(
                    f"Failed to efficiently execute query, but inefficient code path succeeded. Original error was: {efficient_execution_traceback}"
                )
            if len(batch) > 0 or query_execution_parameters.yield_empty_batches:
                yield batch

    async def _async_execute_query_inefficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_feature: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection],
        query_execution_parameters: QueryExecutionParameters,
    ) -> pa.RecordBatch:
        if connection is None:
            try:
                eng = self.get_async_engine()
            except NotImplementedError:
                return await asyncio.get_running_loop().run_in_executor(
                    None,
                    self._execute_query_inefficient,
                    finalized_query,
                    columns_to_feature,
                    connection,
                    query_execution_parameters,
                )
            cnx_ctx = eng.connect()
        else:
            cnx_ctx = async_null_context(connection)
        async with cnx_ctx as cnx:
            async with contextlib.AsyncExitStack() as exit_stack:
                for (
                    _,
                    temp_value,
                    create_temp_table,
                    temp_table,
                    drop_temp_table,
                ) in finalized_query.temp_tables.values():
                    exit_stack.enter_context(
                        self._async_create_temp_table(  # pyright: ignore -- type stubs are wrong
                            create_temp_table, temp_table, drop_temp_table, cnx, temp_value
                        )
                    )
            res = await cnx.execute(finalized_query.query, finalized_query.params)
            desc = res.cursor.description  # type: ignore
            result_columns: list[str] = [col[0] for col in desc]
            features = columns_to_feature(result_columns)
            data: Dict[str, List[Any]] = {}
            for v in features.values():
                # Create an entry for the columns, so the pyarrow table will have the correct columns even
                # if there is no data
                data[v.root_fqn] = []
            for row in res.all():
                for k, v in zip(result_columns, row):
                    if k not in features:
                        # We are not interested in this column
                        continue
                    data[features[k].root_fqn].append(self.convert_db_types(v, features[k].converter))
            # Only keep the columns provided by the schema
            schema = pa.schema([pa.field(v.root_fqn, v.converter.pyarrow_dtype) for v in features.values()])
            return _convert_to_record_batch(data, schema=schema)

    def _execute_query_inefficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[Connection],
        query_execution_parameters: QueryExecutionParameters,
    ) -> pa.RecordBatch:
        with (
            self.get_engine().connect() if connection is None else contextlib.nullcontext(connection) as cnx,
            cnx.begin(),
        ):
            with contextlib.ExitStack() as exit_stack:
                for (
                    _,
                    temp_value,
                    create_temp_table,
                    temp_table,
                    drop_temp_table,
                ) in finalized_query.temp_tables.values():
                    exit_stack.enter_context(
                        self._create_temp_table(create_temp_table, temp_table, drop_temp_table, cnx, temp_value)
                    )
                # TODO: postgres may have timestamp casting errors due to timezones here
                res = cnx.execute(finalized_query.query, finalized_query.params)
                desc = res.cursor.description
                result_columns: list[str] = [col[0] for col in desc]
                features = columns_to_features(result_columns)
                data: Dict[str, List[Any]] = {}
                for v in features.values():
                    # Create an entry for the columns, so the pyarrow table will have the correct columns even
                    # if there is no data
                    data[v.root_fqn] = []
                for row in res.all():
                    for k, v in zip(result_columns, row):
                        if k not in features:
                            # We are not interested in this column
                            continue
                        data[features[k].root_fqn].append(self.convert_db_types(v, features[k].converter))
                # Only keep the columns provided by the schema
                schema = pa.schema([pa.field(v.root_fqn, v.converter.pyarrow_dtype) for v in features.values()])
        return _convert_to_record_batch(data, schema=schema)

    def _execute_query_efficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[Connection],
        query_execution_parameters: QueryExecutionParameters,
    ) -> Iterable[pa.RecordBatch]:
        raise NotImplementedError

    def _async_execute_query_efficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection],
        query_execution_parameters: QueryExecutionParameters,
    ) -> AsyncIterable[pa.RecordBatch]:
        raise NotImplementedError

    def get_sqlalchemy_dialect(self, paramstyle: Optional[str] = None) -> Dialect:
        return self.local_engine_url().get_dialect()(paramstyle=paramstyle)

    def compile_query(
        self,
        finalized_query: FinalizedChalkQuery,
        paramstyle: Optional[str] = None,
    ) -> Tuple[str, Sequence[Any], Dict[str, Any]]:
        """Compile a query into a string and the bindparams"""
        compiled_query = self._get_compiled_query(finalized_query, paramstyle)
        query_string = compiled_query.string
        return query_string, compiled_query.positiontup or [], compiled_query.params or {}

    def _get_compiled_query(
        self,
        finalized_query: FinalizedChalkQuery,
        paramstyle: Optional[str] = None,
    ) -> SQLCompiler:
        """Compile a query into a string and the bindparams"""
        dialect = self.get_sqlalchemy_dialect(paramstyle=paramstyle)
        query = finalized_query.query.params(finalized_query.params)
        compiled_query = query.compile(dialect=dialect)
        return compiled_query

    def _load_env_engine_args(self, name: str, override: Optional[Mapping[str, str]]) -> Mapping[str, Any]:
        """
        Loads additional engine arguments from env var "{name}_ENGINE_ARGUMENTS"
        """

        extra_args = load_integration_variable(integration_name=name, name="ENGINE_ARGUMENTS", override=override)
        if extra_args is None:
            return {}
        else:
            extra_args = json.loads(extra_args)
            assert isinstance(extra_args, dict), "ENGINE_ARGUMENTS must be a JSON object"
            return extra_args

    def _check_engine_isolation_level(self):
        isolation_level = self.engine_args.get("isolation_level")
        if isolation_level == "AUTOCOMMIT":
            warnings.warn(
                UserWarning(
                    (
                        f"The SQL engine '{self.name}' is being created with the AUTOCOMMIT transaction isolation level, which helps improve "
                        "performance for SELECT statements by avoiding unnecessary transactions. If a different transaction level is needed for an "
                        "individual connection, use the `execution_options` method when retrieving a connection -- e.g. "
                        "`with get_engine().connect().execution_options(isolation_level='REPEATABLE READ') as cnx: ...`. "
                        "For more information, please see "
                        "https://docs.sqlalchemy.org/en/20/core/connections.html#setting-isolation-level-or-dbapi-autocommit-for-a-connection"
                    )
                )
            )

    def get_engine(self) -> Engine:
        from sqlalchemy.engine import create_engine

        if self._engine is None:
            self.register_sqlalchemy_compiler_overrides()
            self._check_engine_isolation_level()
            self._engine = create_engine(url=self.local_engine_url(), **self.engine_args)
        return self._engine

    def get_async_engine(self):
        from sqlalchemy.ext.asyncio import create_async_engine

        if self._async_engine is None:
            self.register_sqlalchemy_compiler_overrides()
            self._check_engine_isolation_level()
            self._async_engine = create_async_engine(url=self.async_local_engine_url(), **self.async_engine_args)
        return self._async_engine

    def raw_session(self) -> Session:
        from sqlalchemy.orm import Session

        warnings.warn(
            DeprecationWarning(
                (
                    "The method `raw_session()` is deprecated. Instead, please construct a session directly "
                    "from the underlying engine -- for example:"
                    "`from sqlalchemy.orm import Session; with Session(sql_source.get_engine()) as session: ...`"
                )
            )
        )
        return Session(self.get_engine())

    def convert_db_types(self, v: Any, converter: FeatureConverter):
        """
        Overload this if a given DB type needs custom type conversion
        """
        return converter.from_rich_to_primitive(v, missing_value_strategy="default_or_allow")

    def to_json(self) -> Dict[str, str]:
        return {"name": self.name or "", "kind": self.kind.value}

    @classmethod
    def register_sqlalchemy_compiler_overrides(cls):
        """Hook to register SQLAlchemy Compiler overrides. These hooks are registered by default in tests and in the engine"""
        return

    def _recreate_integration_variables(self) -> dict[str, str]:
        raise NotImplementedError


def _extract_features(feature_set: Features, prefix: str, targets: List[Label], features: Dict[str, Feature]):
    import sqlalchemy.sql.functions
    import sqlalchemy.sql.schema
    from sqlalchemy import column
    from sqlalchemy.orm import InstrumentedAttribute

    for f in feature_set.features:
        assert f.attribute_name is not None
        try:
            feature_value = getattr(feature_set, f.attribute_name)
        except AttributeError:
            continue
        if f.is_has_many:
            raise ValueError(
                f"Feature '{prefix}.{f.name}' is a nested has-many feature, which is not supported when querying from SQL"
            )
        root_fqn = f"{prefix}.{f.name}"
        if isinstance(feature_value, Features):
            # Nested sub-features
            _extract_features(feature_value, root_fqn, targets, features)
        elif isinstance(feature_value, str):
            # Treat it as a column name
            features[root_fqn] = Feature.from_root_fqn(root_fqn)
            targets.append(column(feature_value).label(root_fqn))
        elif isinstance(feature_value, (sqlalchemy.sql.functions.GenericFunction, InstrumentedAttribute)):
            features[root_fqn] = Feature.from_root_fqn(root_fqn)
            targets.append(feature_value.label(root_fqn))
        elif isinstance(feature_value, sqlalchemy.sql.schema.Column):
            features[root_fqn] = Feature.from_root_fqn(root_fqn)
            targets.append(feature_value.label(root_fqn))
        else:
            raise TypeError(
                (
                    f"Feature '{root_fqn}' has an unsupported value of type '{type(feature_value)}' for SQL queries. "
                    "All values must be a column, a column name string, a nested Features class, or an SQLAlchemy function"
                )
            )


insert_size_limit = 10_000


def _convert_to_record_batch(
    data: Mapping[str, List[Any]],
    schema: pa.Schema,
) -> pa.RecordBatch:
    """
    Converts the provided dict of fqn --> values into a `pa.RecordBatch`.
    This method will detect validation errors and report which FQN failed to be parsed,
    and the target Pyarrow type, which makes debugging much easier.
    """
    data_arrays: dict[str, pa.Array] = dict()
    for k, v in data.items():
        try:
            data_arrays[k] = pa.array(v, type=schema.field(k).type)
        except Exception as e:
            raise ValueError(
                f"Unable to convert values for feature '{k}' into pyarrow type '{schema.field(k).type}': {e}"
            )
    return pa.RecordBatch.from_pydict(data_arrays, schema=schema)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Union

from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource, SQLSourceKind, TableIngestMixIn
from chalk.sql.protocols import SQLSourceWithTableIngestProtocol
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import sqlalchemy as sa
    from sqlalchemy.engine import URL, Engine

try:
    import sqlalchemy as sa
except ImportError:
    sa = None

if sa is None:
    _supported_sqlalchemy_types_for_pa_csv_querying = ()
else:
    _supported_sqlalchemy_types_for_pa_csv_querying = (
        sa.BigInteger,
        sa.Boolean,
        sa.Float,
        sa.Integer,
        sa.String,
        sa.Text,
        sa.DateTime,
        sa.Date,
        sa.SmallInteger,
        sa.BIGINT,
        sa.BOOLEAN,
        sa.CHAR,
        sa.DATETIME,
        sa.FLOAT,
        sa.INTEGER,
        sa.SMALLINT,
        sa.TEXT,
        sa.TIMESTAMP,
        sa.VARCHAR,
    )

_logger = get_logger(__name__)


_TRINO_HOST_NAME = "TRINO_HOST"
_TRINO_PORT_NAME = "TRINO_PORT"
_TRINO_CATALOG_NAME = "TRINO_CATALOG"
_TRINO_SCHEMA_NAME = "TRINO_SCHEMA"
_TRINO_USER_NAME = "TRINO_USER"
_TRINO_PASSWORD_NAME = "TRINO_PASSWORD"


class TrinoSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIngestProtocol):
    kind = SQLSourceKind.trino

    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
        catalog: Optional[str] = None,
        schema: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
        engine_args: Optional[Dict[str, Any]] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            import trino
        except ImportError:
            raise missing_dependency_exception("chalkpy[trino]")
        del trino
        self.name = name
        self.host = host or load_integration_variable(
            integration_name=name, name=_TRINO_HOST_NAME, override=integration_variable_override
        )

        self.port = (
            int(port)
            if port is not None
            else load_integration_variable(
                integration_name=name, name=_TRINO_PORT_NAME, parser=int, override=integration_variable_override
            )
        )
        self.catalog = catalog or load_integration_variable(
            integration_name=name, name=_TRINO_CATALOG_NAME, override=integration_variable_override
        )
        self.schema = schema or load_integration_variable(
            integration_name=name, name=_TRINO_SCHEMA_NAME, override=integration_variable_override
        )
        self.user = user or load_integration_variable(
            integration_name=name, name=_TRINO_USER_NAME, override=integration_variable_override
        )
        self.password = password or load_integration_variable(
            integration_name=name, name=_TRINO_PASSWORD_NAME, override=integration_variable_override
        )
        self.ingested_tables: Dict[str, Any] = {}

        if engine_args is None:
            engine_args = {}

        if name:
            engine_args_from_ui = self._load_env_engine_args(name, override=integration_variable_override)
            for k, v in engine_args_from_ui.items():
                engine_args.setdefault(k, v)

        chalk_default_engine_args = {
            "session_properties": {"query_max_run_time": "1d"},
        }
        for k, v in chalk_default_engine_args.items():
            engine_args.setdefault(k, v)

        # We set the default isolation level to autocommit since the SQL sources are read-only, and thus
        # transactions are not needed
        # Setting the isolation level on the engine, instead of the connection, avoids
        # a DBAPI statement to reset the transactional level back to the default before returning the
        # connection to the pool
        BaseSQLSource.__init__(self, name=name, engine_args=engine_args, async_engine_args=None)

    def get_sqlglot_dialect(self) -> str | None:
        return "trino"

    def local_engine_url(self) -> URL:
        import trino.sqlalchemy as ts
        from sqlalchemy.engine import make_url

        if self.host is None:
            raise ValueError("Failed to resolve trino required trino host.")

        return make_url(
            ts.URL(
                user=self.user,
                password=self.password,
                # host is checked for None in the constructor
                host=self.host,
                port=self.port,
                catalog=self.catalog,
                schema=self.schema,
            )
        )

    def get_engine(self) -> Engine:
        from sqlalchemy.engine import create_engine

        if self._engine is None:  # pyright: ignore[reportUnnecessaryComparison]
            self.register_sqlalchemy_compiler_overrides()
            self._check_engine_isolation_level()
            self._engine = create_engine(url=self.local_engine_url(), connect_args=self.engine_args)
        return self._engine

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_TRINO_HOST_NAME, self.name, self.host),
                create_integration_variable(_TRINO_PORT_NAME, self.name, self.port),
                create_integration_variable(_TRINO_CATALOG_NAME, self.name, self.catalog),
                create_integration_variable(_TRINO_SCHEMA_NAME, self.name, self.schema),
                create_integration_variable(_TRINO_USER_NAME, self.name, self.user),
                create_integration_variable(_TRINO_PASSWORD_NAME, self.name, self.password),
            ]
            if v is not None
        }



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.aggregate.v1.service_pb2 import (
    GetActiveCronAggregateBackfillsRequest,
    GetActiveCronAggregateBackfillsResponse,
    GetAggregateBackfillJobRequest,
    GetAggregateBackfillJobResponse,
    GetAggregateBackfillJobsRequest,
    GetAggregateBackfillJobsResponse,
    GetAggregatesRequest,
    GetAggregatesResponse,
    GetCronAggregateBackfillRequest,
    GetCronAggregateBackfillResponse,
    PlanAggregateBackfillRequest,
    PlanAggregateBackfillResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class AggregateServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    PlanAggregateBackfill: UnaryUnaryMultiCallable[
        PlanAggregateBackfillRequest,
        PlanAggregateBackfillResponse,
    ]
    """PlanAggregateBackfill determines the estimated resources needed to backfill
    an aggregate.

    This method is a duplicate of the PlanAggregateBackfill method
    in the query_server.proto file. We should remove the query_server.proto method
    and move that request to this service instead.
    buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
    """
    GetAggregates: UnaryUnaryMultiCallable[
        GetAggregatesRequest,
        GetAggregatesResponse,
    ]
    """This method is a duplicate of the PlanAggregateBackfill method
    in the query_server.proto file. We should remove the query_server.proto method
    and move that request to this service instead.
    buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
    """
    GetAggregateBackfillJobs: UnaryUnaryMultiCallable[
        GetAggregateBackfillJobsRequest,
        GetAggregateBackfillJobsResponse,
    ]
    GetAggregateBackfillJob: UnaryUnaryMultiCallable[
        GetAggregateBackfillJobRequest,
        GetAggregateBackfillJobResponse,
    ]
    GetCronAggregateBackfill: UnaryUnaryMultiCallable[
        GetCronAggregateBackfillRequest,
        GetCronAggregateBackfillResponse,
    ]
    GetActiveCronAggregateBackfills: UnaryUnaryMultiCallable[
        GetActiveCronAggregateBackfillsRequest,
        GetActiveCronAggregateBackfillsResponse,
    ]

class AggregateServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def PlanAggregateBackfill(
        self,
        request: PlanAggregateBackfillRequest,
        context: ServicerContext,
    ) -> PlanAggregateBackfillResponse:
        """PlanAggregateBackfill determines the estimated resources needed to backfill
        an aggregate.

        This method is a duplicate of the PlanAggregateBackfill method
        in the query_server.proto file. We should remove the query_server.proto method
        and move that request to this service instead.
        buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
        """
    @abstractmethod
    def GetAggregates(
        self,
        request: GetAggregatesRequest,
        context: ServicerContext,
    ) -> GetAggregatesResponse:
        """This method is a duplicate of the PlanAggregateBackfill method
        in the query_server.proto file. We should remove the query_server.proto method
        and move that request to this service instead.
        buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
        """
    @abstractmethod
    def GetAggregateBackfillJobs(
        self,
        request: GetAggregateBackfillJobsRequest,
        context: ServicerContext,
    ) -> GetAggregateBackfillJobsResponse: ...
    @abstractmethod
    def GetAggregateBackfillJob(
        self,
        request: GetAggregateBackfillJobRequest,
        context: ServicerContext,
    ) -> GetAggregateBackfillJobResponse: ...
    @abstractmethod
    def GetCronAggregateBackfill(
        self,
        request: GetCronAggregateBackfillRequest,
        context: ServicerContext,
    ) -> GetCronAggregateBackfillResponse: ...
    @abstractmethod
    def GetActiveCronAggregateBackfills(
        self,
        request: GetActiveCronAggregateBackfillsRequest,
        context: ServicerContext,
    ) -> GetActiveCronAggregateBackfillsResponse: ...

def add_AggregateServiceServicer_to_server(servicer: AggregateServiceServicer, server: Server) -> None: ...



from __future__ import annotations

import collections.abc
import contextvars
import dataclasses
import inspect
from datetime import datetime, timedelta, timezone
from types import TracebackType
from typing import TYPE_CHECKING, Any, Hashable, Set, Type

if TYPE_CHECKING:
    from chalk.features.feature_field import Feature


class ClauseJoinWithAndException(Exception):
    ...


class Filter:
    def __init__(self, lhs: Any, operation: str, rhs: Any):
        from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature

        super().__init__()

        # Feature or other could be another feature, filter, featuretime, literal
        # Other could also be a sequence (in the case of operation = "in")
        self.operation = operation
        if isinstance(lhs, FeatureWrapper):
            lhs = unwrap_feature(lhs, raise_error=False)
        self.lhs = lhs
        if self.operation == "in":
            if not isinstance(rhs, collections.abc.Iterable):
                raise ValueError("The RHS must be an iterable for operation='in'")
            rhs_unwrapped = [unwrap_feature(x, raise_error=False) if isinstance(x, FeatureWrapper) else x for x in rhs]
            for x in rhs_unwrapped:
                if not isinstance(x, Hashable):
                    raise ValueError(f"RHS for operation='in' must be hashable; Found un-hashable type: {type(x)}")
            rhs = frozenset(unwrap_feature(x, raise_error=False) if isinstance(x, FeatureWrapper) else x for x in rhs)
        else:
            if isinstance(rhs, FeatureWrapper):
                rhs = unwrap_feature(rhs, raise_error=False)
        self.rhs = rhs

    def __hash__(self) -> int:
        return hash((self.lhs, self.operation, self.rhs))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Filter):
            return NotImplemented
        return self.lhs == other.lhs and self.operation == other.operation and self.rhs == other.rhs

    def __and__(self, other: object):
        return Filter(self, "and", other)

    def __or__(self, other: object):
        return Filter(self, "or", other)

    def __invert__(self):
        return Filter(self, "not", None)

    def referenced_features(self) -> Set[Feature]:
        from chalk.features.feature_field import Feature

        return {x for x in (self.lhs, self.rhs) if isinstance(x, Feature)}

    def __repr__(self):
        return f"Filter({self.lhs} {self.operation} {self.rhs})"

    def __bool__(self):
        frame = inspect.currentframe()
        if frame and frame.f_back:
            frame = frame.f_back
            if frame.f_code.co_name == "<lambda>":
                raise ClauseJoinWithAndException()  # must join with '&', not 'and'
        del frame
        return bool(self.rhs == self.lhs)


def _get_feature_time_index(explicit_index: Any):
    if explicit_index is None:
        from chalk.features.pseudofeatures import CHALK_TS_FEATURE

        return CHALK_TS_FEATURE

    from chalk.features.feature_wrapper import unwrap_feature

    return unwrap_feature(explicit_index)


@dataclasses.dataclass(frozen=True)
class TimeDelta:
    weeks_ago: int = 0
    days_ago: int = 0
    hours_ago: int = 0
    minutes_ago: int = 0
    seconds_ago: int = 0
    milliseconds_ago: int = 0
    microseconds_ago: int = 0

    def to_std(self) -> timedelta:
        # Returning the negative inverse since the feature is defined in the past (e.g. days **ago**)
        return -timedelta(
            weeks=self.weeks_ago,
            days=self.days_ago,
            hours=self.hours_ago,
            minutes=self.minutes_ago,
            seconds=self.seconds_ago,
            milliseconds=self.milliseconds_ago,
            microseconds=self.microseconds_ago,
        )


def before(
    weeks_ago: int = 0,
    days_ago: int = 0,
    hours_ago: int = 0,
    minutes_ago: int = 0,
    seconds_ago: int = 0,
    milliseconds_ago: int = 0,
    microseconds_ago: int = 0,
    index: Any = None,
) -> Any:
    """The function `before` can be used with `DataFrame` to compute windowed features.

    `before` filters a `DataFrame` relative to the current time in context such that
    if the `before` filter is defined as `now - {time_window}`, the filter will include
    all features with timestamps t where `t <= now - {time_window}`.
    This time could be in the past if you’re using an offline resolver.
    Using window functions ensures that you maintain point-in-time correctness.

    The parameters to `before` take many keyword arguments describing the
    time relative to the present.

    Parameters
    ----------
    days_ago
        Number of days ago.
    hours_ago
        Number of hours ago.
    minutes_ago
        Number of minutes ago.
    seconds_ago
        Number of seconds ago.
    index
        The feature to use for the filter. By default, `index` is the `FeatureTime`
        of the referenced feature class.

    Other Parameters
    ----------------
    weeks_ago
        Number of weeks ago.
    milliseconds_ago
        Number of milliseconds ago.
    microseconds_ago
        Number of microseconds ago.

    Returns
    -------
    Any
        A filter for a `DataFrame`.

    Examples
    --------
    >>> from chalk.features import DataFrame, features
    >>> @features
    ... class Card:
    ...     ...
    >>> @features
    ... class User:
    ...     cards: DataFrame[Card]
    >>> User.cards[before(hours_ago=1, minutes_ago=30)]
    """
    return Filter(
        lhs=_get_feature_time_index(index),
        operation="<=",
        rhs=TimeDelta(
            weeks_ago=weeks_ago,
            days_ago=days_ago,
            hours_ago=hours_ago,
            minutes_ago=minutes_ago,
            seconds_ago=seconds_ago,
            milliseconds_ago=milliseconds_ago,
            microseconds_ago=microseconds_ago,
        ),
    )


def after(
    weeks_ago: int = 0,
    days_ago: int = 0,
    hours_ago: int = 0,
    minutes_ago: int = 0,
    seconds_ago: int = 0,
    milliseconds_ago: int = 0,
    microseconds_ago: int = 0,
    index: Any = None,
) -> Any:
    """The function `after` can be used with `DataFrame` to compute windowed features.

    `after` filters a `DataFrame` relative to the current time in context, such that if
    the `after` filter is defined as `now - {time_window}`, the filter will include
    all features with timestamps t where `now - {time_window} <= t <= now`.
    This time could be in the past if you’re using an offline resolver.
    Using window functions ensures that you maintain point-in-time correctness.

    The parameters to `after` take many keyword arguments describing the
    time relative to the present.

    Parameters
    ----------
    days_ago
        Number of days ago.
    hours_ago
        Number of hours ago.
    minutes_ago
        Number of minutes ago.
    seconds_ago
        Number of seconds ago.
    index
        The feature to use for the filter. By default, `index` is the `FeatureTime`
        of the referenced feature class.

    Other Parameters
    ----------------
    weeks_ago
        Number of weeks ago.
    milliseconds_ago
        Number of milliseconds ago.
    microseconds_ago
        Number of microseconds ago.

    Returns
    -------
    Any
        A filter for the `DataFrame`.

    Examples
    --------
    >>> from chalk.features import DataFrame, features
    >>> @features
    ... class Card:
    ...     ...
    >>> @features
    ... class User:
    ...     cards: DataFrame[Card]
    >>> User.cards[after(hours_ago=1, minutes_ago=30)]
    """
    return Filter(
        lhs=_get_feature_time_index(index),
        operation=">=",
        rhs=TimeDelta(
            weeks_ago=weeks_ago,
            days_ago=days_ago,
            hours_ago=hours_ago,
            minutes_ago=minutes_ago,
            seconds_ago=seconds_ago,
            milliseconds_ago=milliseconds_ago,
            microseconds_ago=microseconds_ago,
        ),
    )


_FILTER_NOW_CONTEXT_STACK: contextvars.ContextVar[datetime] = contextvars.ContextVar("_FILTER_NOW_CONTEXT_STACK")


def get_filter_now() -> datetime:
    return _FILTER_NOW_CONTEXT_STACK.get(datetime.now(tz=timezone.utc))


def time_is_frozen() -> bool:
    rt = datetime.now(tz=timezone.utc)
    now = _FILTER_NOW_CONTEXT_STACK.get(rt)
    return now != rt


class freeze_time:
    def __init__(self, at: datetime):
        """
        Used to freeze the 'now' value used to execute filters like `after(days_ago=30)`.

        Parameters
        ----------
        at
            The time to freeze to. Must be timezone aware.

        Examples
        --------
        >>> from chalk.features import online, DataFrame
        >>> from datetime import datetime, timedelta, timezone
        >>> @online
        ... def get_average_spend_30d(
        ...     spend: User.cards[after(days_ago=30)],
        ... ) -> User.average_spend_30d:
        ...     return spend.mean()
        >>> with freeze_time(datetime(2021, 1, 1, tzinfo=timezone.utc)):
        >>>     now = datetime.now(tz=timezone.utc)
        >>>     get_average_spend_30d(
        ...         spend=DataFrame([
        ...             Card(spend=10, ts=now - timedelta(days=31)),
        ...             Card(spend=20, ts=now - timedelta(days=29)),
        ...             Card(spend=30, ts=now - timedelta(days=28)),
        ...         ])
        ...     )
        """
        super().__init__()
        if at.tzinfo is None:
            raise ValueError("Timestamps used with 'freeze_time' must have timezone information associated with them.")
        self._time = at
        # Keep a list of previous contextvar tokens so that this context manager is re-entrant
        # (i.e. you can enter multiple times and then exit multiple times)
        self._token_stack = []

    def time(self) -> datetime:
        """
        Returns the current time that filters will use.

        Returns
        -------
        datetime
            The current time that filters will use.

        Examples
        --------
        >>> with freeze_time(datetime(2021, 1, 1, tzinfo=timezone.utc)) as ft:
        ...     assert ft.time() == datetime(2021, 1, 1, tzinfo=timezone.utc)
        """
        return self._time

    def __enter__(self):
        """The freeze_time class is a context manager, so it can be used with the `with` statement.
        The `__enter__` method is called when entering the context manager.
        """
        token = _FILTER_NOW_CONTEXT_STACK.set(self._time)
        self._token_stack.append(token)
        return self

    def __exit__(
        self,
        __exc_type: Type[BaseException] | None,
        __exc_value: BaseException | None,
        __traceback: TracebackType | None,
    ) -> bool | None:
        """
        The freeze_time class is a context manager, so it can be used with the `with` statement.
        The `__exit__` method is automatically called when exiting the context manager.
        """
        previous_token = self._token_stack.pop(-1)
        _FILTER_NOW_CONTEXT_STACK.reset(previous_token)
        return None


__all__ = [
    "Filter",
    "TimeDelta",
    "after",
    "before",
]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.http_engine_proxy_pb2 import (
    CreateOfflineQueryRequest,
    CreateOfflineQueryResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class HTTPEngineProxyServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    CreateOfflineQuery: UnaryUnaryMultiCallable[
        CreateOfflineQueryRequest,
        CreateOfflineQueryResponse,
    ]

class HTTPEngineProxyServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def CreateOfflineQuery(
        self,
        request: CreateOfflineQueryRequest,
        context: ServicerContext,
    ) -> CreateOfflineQueryResponse: ...

def add_HTTPEngineProxyServiceServicer_to_server(servicer: HTTPEngineProxyServiceServicer, server: Server) -> None: ...



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.kubernetes.v1 import events_pb2 as _events_pb2
from chalk._gen.chalk.kubernetes.v1 import persistentvolume_pb2 as _persistentvolume_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class GetPodStackTraceDumpRequest(_message.Message):
    __slots__ = ("namespace", "pod_name", "container_name", "process_id", "process_name", "auto_detect_process")
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    POD_NAME_FIELD_NUMBER: _ClassVar[int]
    CONTAINER_NAME_FIELD_NUMBER: _ClassVar[int]
    PROCESS_ID_FIELD_NUMBER: _ClassVar[int]
    PROCESS_NAME_FIELD_NUMBER: _ClassVar[int]
    AUTO_DETECT_PROCESS_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    pod_name: str
    container_name: str
    process_id: int
    process_name: str
    auto_detect_process: bool
    def __init__(
        self,
        namespace: _Optional[str] = ...,
        pod_name: _Optional[str] = ...,
        container_name: _Optional[str] = ...,
        process_id: _Optional[int] = ...,
        process_name: _Optional[str] = ...,
        auto_detect_process: bool = ...,
    ) -> None: ...

class GetPodStackTraceDumpResponse(_message.Message):
    __slots__ = ("stack_trace",)
    STACK_TRACE_FIELD_NUMBER: _ClassVar[int]
    stack_trace: str
    def __init__(self, stack_trace: _Optional[str] = ...) -> None: ...

class GetKubernetesEventsRequest(_message.Message):
    __slots__ = ("cluster_name", "namespace", "label_selector", "field_selector")
    CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    LABEL_SELECTOR_FIELD_NUMBER: _ClassVar[int]
    FIELD_SELECTOR_FIELD_NUMBER: _ClassVar[int]
    cluster_name: str
    namespace: str
    label_selector: str
    field_selector: str
    def __init__(
        self,
        cluster_name: _Optional[str] = ...,
        namespace: _Optional[str] = ...,
        label_selector: _Optional[str] = ...,
        field_selector: _Optional[str] = ...,
    ) -> None: ...

class GetKubernetesEventsResponse(_message.Message):
    __slots__ = ("events",)
    EVENTS_FIELD_NUMBER: _ClassVar[int]
    events: _containers.RepeatedCompositeFieldContainer[_events_pb2.ChalkKubernetesEvent]
    def __init__(
        self, events: _Optional[_Iterable[_Union[_events_pb2.ChalkKubernetesEvent, _Mapping]]] = ...
    ) -> None: ...

class GetKubernetesPersistentVolumesRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetKubernetesPersistentVolumesResponse(_message.Message):
    __slots__ = ("volumes",)
    VOLUMES_FIELD_NUMBER: _ClassVar[int]
    volumes: _containers.RepeatedCompositeFieldContainer[_persistentvolume_pb2.ChalkKubernetesPersistentVolume]
    def __init__(
        self,
        volumes: _Optional[_Iterable[_Union[_persistentvolume_pb2.ChalkKubernetesPersistentVolume, _Mapping]]] = ...,
    ) -> None: ...



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.kubernetes.v1 import nodes_pb2 as _nodes_pb2
from chalk._gen.chalk.kubernetes.v1 import pods_pb2 as _pods_pb2
from chalk._gen.chalk.pubsub.v1 import node_status_pb2 as _node_status_pb2
from chalk._gen.chalk.pubsub.v1 import pod_status_pb2 as _pod_status_pb2
from chalk._gen.chalk.server.v1 import chart_pb2 as _chart_pb2
from chalk._gen.chalk.server.v1 import pod_request_pb2 as _pod_request_pb2
from chalk._gen.chalk.usage.v1 import rate_pb2 as _rate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class UsageChartPeriod(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    USAGE_CHART_PERIOD_UNSPECIFIED: _ClassVar[UsageChartPeriod]
    USAGE_CHART_PERIOD_DAILY: _ClassVar[UsageChartPeriod]
    USAGE_CHART_PERIOD_MONTHLY: _ClassVar[UsageChartPeriod]

class UsageChartGrouping(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    USAGE_CHART_GROUPING_UNSPECIFIED: _ClassVar[UsageChartGrouping]
    USAGE_CHART_GROUPING_INSTANCE_TYPE: _ClassVar[UsageChartGrouping]
    USAGE_CHART_GROUPING_CLUSTER: _ClassVar[UsageChartGrouping]

USAGE_CHART_PERIOD_UNSPECIFIED: UsageChartPeriod
USAGE_CHART_PERIOD_DAILY: UsageChartPeriod
USAGE_CHART_PERIOD_MONTHLY: UsageChartPeriod
USAGE_CHART_GROUPING_UNSPECIFIED: UsageChartGrouping
USAGE_CHART_GROUPING_INSTANCE_TYPE: UsageChartGrouping
USAGE_CHART_GROUPING_CLUSTER: UsageChartGrouping

class GetUsageChartRequest(_message.Message):
    __slots__ = ("start_ms", "end_ms", "period", "grouping")
    START_MS_FIELD_NUMBER: _ClassVar[int]
    END_MS_FIELD_NUMBER: _ClassVar[int]
    PERIOD_FIELD_NUMBER: _ClassVar[int]
    GROUPING_FIELD_NUMBER: _ClassVar[int]
    start_ms: int
    end_ms: int
    period: UsageChartPeriod
    grouping: UsageChartGrouping
    def __init__(
        self,
        start_ms: _Optional[int] = ...,
        end_ms: _Optional[int] = ...,
        period: _Optional[_Union[UsageChartPeriod, str]] = ...,
        grouping: _Optional[_Union[UsageChartGrouping, str]] = ...,
    ) -> None: ...

class GetUsageChartResponse(_message.Message):
    __slots__ = ("chart",)
    CHART_FIELD_NUMBER: _ClassVar[int]
    chart: _chart_pb2.Chart
    def __init__(self, chart: _Optional[_Union[_chart_pb2.Chart, _Mapping]] = ...) -> None: ...

class GetUtilizationRatesRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetUtilizationRatesResponse(_message.Message):
    __slots__ = ("rates",)
    RATES_FIELD_NUMBER: _ClassVar[int]
    rates: _containers.RepeatedCompositeFieldContainer[_rate_pb2.MachineRate]
    def __init__(self, rates: _Optional[_Iterable[_Union[_rate_pb2.MachineRate, _Mapping]]] = ...) -> None: ...

class GetNodesAndPodsRequest(_message.Message):
    __slots__ = ("namespace", "pod_label_selector", "environment_id")
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    POD_LABEL_SELECTOR_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    pod_label_selector: str
    environment_id: str
    def __init__(
        self,
        namespace: _Optional[str] = ...,
        pod_label_selector: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
    ) -> None: ...

class GetNodesAndPodsResponse(_message.Message):
    __slots__ = ("nodes", "pods")
    NODES_FIELD_NUMBER: _ClassVar[int]
    PODS_FIELD_NUMBER: _ClassVar[int]
    nodes: _containers.RepeatedCompositeFieldContainer[_node_status_pb2.NodeStatusPubSub]
    pods: _containers.RepeatedCompositeFieldContainer[_pod_status_pb2.PodStatusPubSub]
    def __init__(
        self,
        nodes: _Optional[_Iterable[_Union[_node_status_pb2.NodeStatusPubSub, _Mapping]]] = ...,
        pods: _Optional[_Iterable[_Union[_pod_status_pb2.PodStatusPubSub, _Mapping]]] = ...,
    ) -> None: ...

class GetNodesAndPodsUIRequest(_message.Message):
    __slots__ = ("namespace", "pod_label_selector", "environment_id")
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    POD_LABEL_SELECTOR_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    pod_label_selector: str
    environment_id: str
    def __init__(
        self,
        namespace: _Optional[str] = ...,
        pod_label_selector: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
    ) -> None: ...

class GetNodesAndPodsUIResponse(_message.Message):
    __slots__ = ("nodes", "pods")
    NODES_FIELD_NUMBER: _ClassVar[int]
    PODS_FIELD_NUMBER: _ClassVar[int]
    nodes: _containers.RepeatedCompositeFieldContainer[_nodes_pb2.KubernetesNodeData]
    pods: _containers.RepeatedCompositeFieldContainer[_pods_pb2.KubernetesPodData]
    def __init__(
        self,
        nodes: _Optional[_Iterable[_Union[_nodes_pb2.KubernetesNodeData, _Mapping]]] = ...,
        pods: _Optional[_Iterable[_Union[_pods_pb2.KubernetesPodData, _Mapping]]] = ...,
    ) -> None: ...

class SyncUtilizationRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SyncUtilizationResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...
